#==========================================================
#
#  Chris Nelson, 2025-
#
#==========================================================


from .core import logging

import signal
import time
import os
import sys
import multiprocessing
import subprocess
import traceback

# Logging events within this module are at the DEBUG level.  With this module's child logger set to
# a minimum of WARNING level by default, then logging from this module is effectively disabled.  To enable
# logging from this module add this within your tool script code:
#       logging.getLogger('cjnfuncs.rwt').setLevel(logging.DEBUG)
rwt_logger = logging.getLogger('cjnfuncs.rwt')
rwt_logger.setLevel(logging.WARNING)


def run_with_timeout(func, *args, **kwargs):
    pass
    """
## run_with_timeout (func, *args, **kwargs, rwt_timeout=1.0, rwt_ntries=1, rwt_kill=True) - Run a function in a separate process with an enforced timeout.

`run_with_timeout` uses the multiprocessing module, and works by running the specified `func` in a managed 
external process that can be reliably killed on timeout.
For a non-timeout run, `run_with_timeout` returns what the `func` returns or any exception raised. 
On timeout, the process is killed (by default) and a TimeoutError exception is raised.


### Args

`func` (callable)
- The function to run
- May be any function - built-in, standard library, supplied by an installed package, or user-written
- On Windows `func` must be defined as a top level function of the module

`*args` (0+)
- Positional args required by func

`**kwargs` (0+)
- Keyword args to be passed to func

`rwt_timeout` additional kwarg (int or float, default 1.0)
- Enforced timeout in seconds

`rwt_ntries` additional kwarg (int, default 1)
- Number of attempts to run `func` if rwt_timeout is exceeded or `func` raises an exception

`rwt_kill` additional kwarg (bool, default True)
- If True, on timeout kill the process
- If False, on timeout let the process continue to run.  It will be orphaned - see Behavior notes, below.


### Returns
- With no timeout or exception, returns the value returned from `func`
- Any exception raised by `func`
- If rwt_timeout is exceeded, returns TimeoutError
- Exceptions raised for invalid rwt_timeout, rwt_ntries, or rwt_kill values


### Behaviors and rules
- Logging within the called `func` is done at the logging level in effect when run_with_timeout is called. 
`logging.getLogger('cjnfuncs.rwt').setLevel(logging.DEBUG)` enables additional status and trace info, intended for debug and regression testing.
- If making a subprocess call and the subprocess timeout limit is triggered, a
subprocess.TimeoutExpired exception is produce with an odd error message on Python 3.11.9: 
`TypeError: TimeoutExpired.__init__() missing 1 required positional argument: 'timeout'`. Generally, don't use
the subprocess timeout arg when using run_with_timeout.
- If `rwt_kill=False` then the forked/spawned process will not be killed, and if the process doesn't exit by itself 
then the tool script will hang on exit, waiting for the orphan process to terminate.
To solve this the tool script needs to kill any orphaned processes created by run_with_timeout before exiting. 
The pids of the orphaned processes are listed in the TimeoutError exception when `rwt_kill=False`, and can
be captured for explicitly killing of any unterminated orphaned processes before exiting the tool script, eg: 
`os.kill (pid, signal.OSKILL)`.  See `rwt.md` for a working example.
Note that if `rwt_ntries` is greater than 1 and `rwt_kill=False`, then potentially several processes may 
be created and orphaned, all attempting to doing the same work.
- On Windows, debug logging messages from the run_with_timeout internal `worker()` function (which calls `func`) are 
erratically produced, and not produced if `func` raises an exception.  Logging from within `func` can also be
erratic.  Raised exceptions operate normally.
"""

    #--------- Top_level ---------
def run_with_timeout(func, *args, **kwargs):

    _timeout = 1.0
    if 'rwt_timeout' in kwargs:
        _timeout = kwargs['rwt_timeout']
        del kwargs['rwt_timeout']
        if not isinstance(_timeout, (int, float)):
            raise ValueError (f"rwt_timeout must be type int or float, received <{_timeout}>")

    _ntries = 1
    if 'rwt_ntries' in kwargs:
        _ntries = kwargs['rwt_ntries']
        del kwargs['rwt_ntries']
        if not isinstance(_ntries, (int)):
            raise ValueError (f"rwt_ntries must be type int, received <{_ntries}>")

    _kill = True
    if 'rwt_kill' in kwargs:
        _kill = kwargs['rwt_kill']
        del kwargs['rwt_kill']
        if not isinstance(_kill, bool):
            raise ValueError (f"rwt_kill must be type bool, received <{_kill}>")
    if _kill == False:
        pid_list = []

    xx =  f"\nrun_with_timeout switches:\n  rwt_timeout:  {_timeout}\n  rwt_ntries:   {_ntries}\n  rwt_kill:     {_kill}"
    xx += f"\n  Function:     {func}\n  args:         {args}\n  kwargs:       {kwargs}"
    rwt_logger.debug (xx)


    kwargs['rwt_loglevel'] = rwt_logger.level
    kwargs['root_loglevel'] = logging.getLogger().level

    for ntry in range(_ntries):

        if _ntries > 1:
            rwt_logger.debug (f"T0  - Try {ntry}")

        # Run it
        rwt_logger.debug (f"T1  - Starting worker_p")
        worker_to_toplevel_q = multiprocessing.get_context().Queue()
        worker_p = multiprocessing.Process(target=worker, args=(worker_to_toplevel_q, func, args, kwargs), daemon=False, name=f'rwt_{func}')
        worker_p.start()
        worker_p.join(timeout=_timeout)
        status=None

        # Kill it if it did not exit normally
        if worker_p.is_alive():
            if _kill:                                       # worker_p is alive.  Kill it.
                rwt_logger.debug (f"T4  - terminate worker_p")
                worker_p.terminate()
                worker_p.join(timeout=1) #0.2)
                if worker_p.is_alive():
                    try:
                        if sys.platform.startswith("win"):
                            rwt_logger.debug (f"T5  - taskkill worker_p pid {worker_p.pid}")
                            subprocess.run(["taskkill", "/PID", str(worker_p.pid), "/F"])
                        else:   # Linux
                            rwt_logger.debug (f"T5  - SIGKILL worker_p pid {worker_p.pid}")
                            os.kill(worker_p.pid, signal.SIGKILL)
                    except Exception as e:
                        if 'No such process' in str(e):     # Corner case of worker_p either ended normally, or the terminate finally happened
                            pass
                        else:
                            if ntry == _ntries-1:
                                raise
                if ntry == _ntries-1:
                    raise TimeoutError (f"Function <{func.__name__}> timed out after {_timeout} seconds (killed)")
            else:                                           # worker_p is alive, and DON'T kill it
                pid_list.append(str(worker_p.pid))
                if ntry == _ntries-1:
                    raise TimeoutError (f"Function <{func.__name__}> timed out after {_timeout} seconds (not killed) orphaned pids: {' '.join(pid_list)}")
        else:
            rwt_logger.debug (f"T2  - worker_p exited before rwt_timeout")

            # Get returned result or exception
            try:
                status, payload = worker_to_toplevel_q.get(timeout=0.05)
            except Exception as e:
                rwt_logger.debug (f"T7  - Nothing returned from runner")
                status = None

        if status:
            rwt_logger.debug (f"T3  - <{status}> msg received from worker_p")
            if status == "result":
                return payload
            elif status == "exception":
                if ntry == _ntries-1:
                    ex_type, ex_msg, ex_trace = payload     # ex_trace retained for possible future debug/use
                    raise ex_type(f"{ex_msg}")


def worker(result_q, func, args, kwargs):

    def worker_int_handler(sig, frame):
        rwt_logger.debug(f"WH1 - Signal {sig} received")
        time.sleep(0.01)                                    # allow time for logging before terminating
        sys.exit()
    signal.signal(signal.SIGTERM, worker_int_handler)       # kill (15)

    logging.getLogger().setLevel(kwargs['root_loglevel'])   # Necessary on Windows since executed func is in the spawned process (defaults to logging.WARNING)
    del kwargs['root_loglevel']

    rwt_logger = logging.getLogger('cjnfuncs.rwt')
    rwt_logger.setLevel(kwargs['rwt_loglevel'])             # Necessary on Windows since worker is in the spawned process (defaults to logging.WARNING)
    del kwargs['rwt_loglevel']

    # Run it
    rwt_logger.debug(f"W1  - worker_p pid {os.getpid()}")

    try:
        result = func(*args, **kwargs)
        result_q.put(("result", result))
    except Exception as e:
        result_q.put(("exception", (e.__class__, str(e), traceback.format_exc())))



#!/usr/bin/env python3
"""cjnfuncs.deployfiles - Push tool-specific files to their proper locations
"""

#==========================================================
#
#  Chris Nelson, 2018-2025
#
#==========================================================

from pathlib import Path
import shutil
import __main__

from .core import logging
from .mungePath import mungePath
import cjnfuncs.core as core

from importlib_resources import files as ir_files

# Logging events within this module are at the INFO level.  With this module's child logger set to
# a minimum of WARNING level by default, then logging from this module is effectively disabled.  To enable
# logging from this module add this within your tool script code:
#       logging.getLogger('cjnfuncs.deployfiles').setLevel(logging.INFO)
deployfiles_logger = logging.getLogger('cjnfuncs.deployfiles')
deployfiles_logger.setLevel(logging.WARNING)


#=====================================================================================
#=====================================================================================
#  d e p l o y _ f i l e s
#=====================================================================================
#=====================================================================================

def deploy_files(files_list, overwrite=False, missing_ok=False):
    """
## deploy_files (files_list, overwrite=False, missing_ok=False) - Install initial tool script files in user or site space

`deploy_files()` is used to install initial setup files (and directory trees) from the installed package (or tool script) 
to the user or site config and data directories. Suggested usage is with the CLI `--setup-user` or `--setup-site` switches.
Distribution files and directory trees are hosted in `<package_dir>/deployment_files/`.

`deploy_files()` accepts a list of dictionaries defining items to be pushed to user or site space.
Each dictionary defines
the `source` file or directory root to be pushed, the `target_dir` for where to push the source, and optionally
the file and directory permissions for the pushed items.  Ownership matches the running user.


### Args
`files_list` (list of dictionaries)
- A list of dictionaries, each specifying a `source` file or directory tree to be copied to a `target_dir`.
  - `source` - Either an individual file or directory tree within and relative to `<package_dir>/deployment_files/`.
    - No wildcard support
    - `source = ''` will create an empty `target_dir`, if not already existing.
  - `target_dir` - A directory target for the pushed `source`.  It is expanded for user and environment vars, 
    and supports these substitutions (per `set_toolname()`):
    - USER_CONFIG_DIR, USER_DATA_DIR, USER_STATE_DIR, USER_CACHE_DIR
    - SITE_CONFIG_DIR, SITE_DATA_DIR
    - CONFIG_DIR, DATA_DIR, STATE_DIR, CACHE_DIR as determined by the existence of site directories (maybe only useful for testing)
    - Also absolute paths
  - `file_stat` - Permissions set on created files (default 0o644) - See Behaviors and rules, below.
  - `dir_stat` - Permissions set on created directories (default 0o755) - See Behaviors and rules, below.

`overwrite` (bool, default False)
- If `overwrite=False` (default) then only missing files/directories will be deployed, with `file_stat`/`dir_stat` applied.
- If `overwrite=True` then all files/directories will be (re)deployed (using the new `file_stat`/`dir_stat`),
potentially overwriting existing data.

`missing_ok` (bool, default False)
- If `missing_ok=True` then a missing `source` file or directory is tolerated (non-fatal).  This feature is used for testing.
- If `missing_ok=False` (default) and the `source` is missing then a FileNotFoundError exception is raised.  Any files deployed before
the exception will be left in place.


### Returns
- None
- Raises various exceptions on failure


### Behaviors and rules
- An empty directory may be deployed, eg: `{"source": "test_dir/emptydir",  "target_dir": "USER_CONFIG_DIR"}`
deploys the source directory named `emptydir` to `$HOME/.config/<toolname>/emptydir`.

- If the `source` points to an individual file, even if within a subdirectory of `deployment_files`,
that file, and not its nested path, will be deployed.
For example, `{"source": "test_dir/x1",  "target_dir": "USER_CONFIG_DIR/subdir"}` 
deploys just the file `x1` to `$HOME/.config/<toolname>/subdir/x1`. 
If you want to deploy a file and retain the source directory structure, then deploy the entire
source subdirectory, eg: `{"source": "test_dir",  "target_dir": "USER_CONFIG_DIR/subdir"}`, which
deploys the entire test_dir directory structure to `$HOME/.config/<toolname>/subdir/test_dir`.

- `dir_stat` is applied to the parent directory where a file is deployed.
When `source` points to a directory tree, `dir_stat` is also applied to
deployed subdirectories _but not the parent directory_ 
(the leaf dir of the `target_dir`). Higher-level created directories are assigned the user-default permissions
(typically 0o755 - umask).

- The permissions on a directory (the `dir_stat`) are set the first time a file
is deployed to that directory.
With `overwrite=False`, for subsequent file deployments to that directory the first `dir_stat` settings are 
retained (the new `dir_stat` setting is disregarded). 
With `overwrite=True`, an existing directory where a file is deployed will be updated 
to the new `dir_stat` value.

- Directory and file permissions on Windows do not support separate permissions for User, Group and Other, 
and Windows ACLs are not currently supported.
When setting permissions on deployed items on Windows, only the User permission is used even on a file share
hosted on Linux.  Eg, a file permission of 0x644 will deploy with permission 0x666 (only the first octet is used).
    """

    default_file_stat = 0o644
    default_dir_stat  = 0o755

    mapping = [
        ["USER_CONFIG_DIR", core.tool.user_config_dir],
        ["USER_DATA_DIR",   core.tool.user_data_dir],
        ["USER_STATE_DIR",  core.tool.user_state_dir],
        ["USER_CACHE_DIR",  core.tool.user_cache_dir],
        ["SITE_CONFIG_DIR", core.tool.site_config_dir],
        ["SITE_DATA_DIR",   core.tool.site_data_dir],
        ["CONFIG_DIR",      core.tool.config_dir],
        ["DATA_DIR",        core.tool.data_dir],
        ["STATE_DIR",       core.tool.state_dir],
        ["CACHE_DIR     ",  core.tool.cache_dir],
        ]

    def resolve_target(_targ, mkdir=False):
        """Do any CONFIG/DATA replacements.  Return a pathlib full path.
        """
        base_path = ""
        _targ = str(_targ)              # Accept str and pathlib
        for remap in mapping:
            if remap[0] in _targ:
                _targ = _targ.replace(remap[0], "")
                if len(_targ) > 0:
                    _targ = _targ[1:]   # Drops leading '/' after remap removed.
                base_path = remap[1]
                break
        return mungePath(_targ, base_path, mkdir=mkdir).full_path


    def copytree(src_dir, dst_dir, overwrite, file_stat, dir_stat):
        """ Recursively deploy the contents of src_dir to dst_dir
        dst_dir is created by the level above, before calling copytree
        """

        for item in list(src_dir.iterdir()):
            out_item = dst_dir / item.name

            if item.is_dir():
                didnt_exist = False
                if not out_item.exists():
                    didnt_exist = True
                    out_item.mkdir(parents=True)
                    deployfiles_logger.info (f"Created   {out_item}")
                if didnt_exist or overwrite:
                    out_item.chmod(dir_stat)
                copytree(item, out_item, overwrite=overwrite, file_stat=file_stat, dir_stat=dir_stat)

            else:   # is_file
                if not out_item.exists()  or  overwrite:
                    shutil.copy2(item, out_item)
                    if file_stat:
                        out_item.chmod(file_stat)
                    deployfiles_logger.info (f"Deployed  {out_item}")
                else:
                    deployfiles_logger.info (f"File <{out_item}> already exists.  Skipped.")


    if core.tool.main_module.__name__ == "__main__":    # Caller is a tool script file, not an installed module
        my_resources = mungePath(__main__.__file__).parent / "deployment_files"
        # print (f"Script case:  <{my_resources}>")
    else:                                               # Caller is an installed module
        my_resources = ir_files(core.tool.main_module).joinpath("deployment_files")
        # print (f"Module case:  <{my_resources}>")


    # ***** Start of iteration thru dictionaries *****
    for item in files_list:
        file_stat=  item.get("file_stat", default_file_stat)
        dir_stat=   item.get("dir_stat",  default_dir_stat)
        source = item["source"]
        if source != '':
            source =    Path(my_resources.joinpath(item["source"]))

        if source == ''  or  source.is_file():
            target_dir = resolve_target(item["target_dir"])
            didnt_exist = False
            if not target_dir.exists():         # TODO hang risk
                didnt_exist = True
                target_dir.mkdir(parents=True)
                deployfiles_logger.info (f"Created   {target_dir}")
            if didnt_exist or overwrite:
                target_dir.chmod(dir_stat)

            if source != '':
                outfile = target_dir / source.name
                if not outfile.exists()  or  overwrite:
                    shutil.copy2 (source, outfile)
                    outfile.chmod(file_stat)
                    deployfiles_logger.info (f"Deployed  {outfile}")
                else:
                    deployfiles_logger.info (f"File <{outfile}> already exists.  Skipped.")

        elif source.is_dir():
            # TODO ONLY WORKS if the source dir is on the file system (eg, not in a package .zip) ????

            target_dir = resolve_target(item["target_dir"]) / source.name

            didnt_exist = False
            if not target_dir.exists():
                didnt_exist = True
                target_dir.mkdir(parents=True)
                deployfiles_logger.info (f"Created   {target_dir}")
            if didnt_exist or overwrite:
                target_dir.chmod(dir_stat)

            copytree(source, target_dir, overwrite=overwrite, file_stat=file_stat, dir_stat=dir_stat)

        elif missing_ok:
            deployfiles_logger.info (f"Can't deploy source <{source.name}>.  Item not found and missing_ok=True.  Skipping.")
        
        else:
            raise FileNotFoundError (f"Can't deploy <{source.name}>.  Item not found.")



#!/usr/bin/env python3
"""cjnfuncs.mungePath - pathlib Paths made easy and useful
"""

#==========================================================
#
#  Chris Nelson, 2018-2025
#
#==========================================================

import os.path
from pathlib import Path, PurePath
from .core import logging
from .rwt import run_with_timeout


#=====================================================================================
#=====================================================================================
#  C l a s s   m u n g e P a t h
#=====================================================================================
#=====================================================================================

class mungePath():
    def __init__(self, in_path='', base_path='', mkdir=False, set_attributes=False, timeout=1.0, ntries=1):
        """
## Class mungePath (in_path='', base_path='', mkdir=False, set_attributes=False, timeout=1.0, ntries=1) - A clean interface for dealing with filesystem paths

`mungePath()` is based on pathlib, producing Path type attributes and status booleans which may be used with all
pathlib.Path methods, such as .open().  `mungePath()` accepts paths in two parts - the tool script specific
portion `in_path` and a `base_path` (prepended if `in_path` is relative), and returns an instance that may 
be cleanly used in the tool script code.
User (`~user/`) and environment vars (`$HOME/`) are supported and expanded.


### Args
`in_path` (Path or str, default '')
- An absolute or relative path to a file or directory, such as `mydir/myfile.txt`
- If `in_path` is an absolute path then the `base_path` is disregarded.
- If `in_path` starts with `./` then the absolute path to the current working directory (cwd) is prepended 
to `in_path`, and the `base_path` is disregarded.  See Special handling note, below.

`base_path` (Path or str, default '')
- An absolute or relative path to a directory, such as `~/.config/mytool`
- `base_path` is prepended to `in_path` if `in_path` is a relative path
- `base_path = ''` (the default) results in a relative path based on the shell current working directory (cwd)
- `base_path = '.'` results in an absolute path based on the shell cwd
- `base_path = core.tool.main_dir` results in an absolute path based on the tool script directory

`mkdir` (bool, default False)
- Force-make a full directory path.  `base_path` / `in_path` is understood to be to a directory.

`set_attributes` (bool, default False) - _See the first note in Behaviors and rules, below_
- If True then `refresh_stats()` is called, setting `.exists`, `.is_file`, `.is_dir` to valid values (or False on timeout).
- If False then those attributes are set to `None`, indicating not initialized.
- These other attributes are always set: `.full_path`, `.parent`, `.name`, `.is_absolute`, and `.is_relative`, as the do not depend on file system access.

`timeout` (int or float, default 1.0 second)
- If `set_attributes=True` then `refresh_stats()` is called using this timeout value for each of the three `run_with_timeout()` calls.

`ntries` (int, default 1)
If `set_attributes=True` then `refresh_stats()` is called using this ntries value for each of the three `run_with_timeout()` calls.


### Returns
- Handle to `mungePath()` instance


### Instance attributes

Attribute | Type | Description
-- | -- | --
`.full_path`     | Path     |   The full expanduser/expandvars path to a file or directory (may not exist)
`.parent`        | Path     |   The directory above the .full_path
`.name`          | str      |   Just the name.suffix of the .full_path
`.is_absolute`   | Boolean  |   True if the .full_path starts from the filesystem root (isn't a relative path) 
`.is_relative`   | Boolean  |   Not .is_absolute
`.exists`        | Boolean  |   True if the .full_path item (file or dir) actually exists **
`.is_file`       | Boolean  |   True if the .full_path item exists and is a file **
`.is_dir`        | Boolean  |   True if the .full_path item exists and is a directory **

** These attributes are set to None, by default.  See the `set_attributes` arg, above, and the first note in Behaviors and rules, below.


### Behaviors and rules
- NOTE: `set_attributes` was added in cjnfuncs version 3.0 and defaults to False, effectively disabling the setting of
`.exists`, `.is_file` and `.is_dir` attributes. When pointing to a remote filesystem that is flaky, evaluating these attributes
can lead to long hang times. If `set_attributes=True` then `refresh_stats()` is called which attempts to populates `.exists`, `.is_file`, 
and `.is_dir` attributes with real values.  On timeout each attribute is set to False.  refresh_stats() uses run_with_timeout() with The default timeout of 1 second is used.
, or will time out with approximately a total 3 second delay.  If these attributes are not needed
for a given mungePath instance then don't set `set_attributes=True`.  The recommended alternatives are to call `check_path_exists()` 
(which uses `run_with_timeout()` with `rwt_ntries=1` and `rwt_timeout=1.0`), or to implement `run_with_timeout()` calls in your code. 

- If `in_path` is a relative path (eg, `mydir/myfile.txt`) portion then the `base_path` is prepended.  
- If both `in_path` and `base_path` are relative then the combined path will also be relative, usually to
the shell cwd.
- If `in_path` is an absolute path (eg, `/tmp/mydir/myfile.txt`) then the `base_path` is disregarded.
- **Special handling for `in_path` starting with `./`:**  Normally, paths starting with `.` are relative paths.
mungePath interprets `in_path` starting with `./` as an absolute path reference to the shell current working 
directory (cwd).
Often in a tool script a user path input is passed to the `in_path` arg.  Using the `./` prefix, a file in 
the shell cwd may be
referenced, eg `./myfile`.  _Covering the cases, assuming the shell cwd is `/home/me`:_

    in_path | base_path | .full_path resolves to
    -- | -- | --
    myfile          | /tmp  | /tmp/myfile
    ./myfile        | /tmp  | /home/me/myfile
    ../myfile       | /tmp  | /tmp/../myfile
    ./../myfile     | /tmp  | /home/me/../myfile
    xyz/myfile      | /tmp  | /tmp/xyz/myfile
    ./xyz/myfile    | /tmp  | /home/me/xyz/myfile

- `in_path` and `base_path` may be type str(), Path(), or PurePath().
- Symlinks are followed (not resolved).
- User and environment vars are expanded, eg `~/.config` >> `/home/me/.config` (`C:\\Users\\me` on Windows), as does `$HOME/.config`. 
Environment var `$HOME` on Linux is equivalent to `%HOMEDRIVE%%HOMEPATH%` on Windows.  mungePath does not
make the substitution since `$HOME` is just one of many possible environment vars.
- The `.parent` is the directory containing (above) the `.full_path`.  If the object `.is_file` then `.parent` is the
directory containing the file.  If the object `.is_dir` then the `.full_path` includes the end-point directory, and 
`.parent` is the directory above the end-point directory.
- When using `mkdir=True` the combined `base_path` / `in_path` is understood to be a directory path (not
to a file), and that directory if possible. (Uses `pathlib.Path.mkdir()`).  A FileExistsError 
is raised if you attempt to mkdir on top of an existing file.
- See [GitHub repo](https://github.com/cjnaz/cjnfuncs) tests/demo-mungePath.py for numerous application examples.
        """

        in_path = str(in_path)
        base_path = str(base_path)

        if in_path.startswith('./'):
            in_path = Path.cwd() / in_path

        in_path_pp = PurePath(os.path.expandvars(os.path.expanduser(str(in_path))))

        if not in_path_pp.is_absolute():
            _base_path = str(base_path)
            if _base_path.startswith("."):
                _base_path = Path.cwd() / _base_path
            _base_path = PurePath(os.path.expandvars(os.path.expanduser(str(_base_path))))
            in_path_pp = _base_path / in_path_pp

        if mkdir:
            Path(in_path_pp).mkdir(parents=True, exist_ok=True)

        self.parent =       Path(in_path_pp.parent)
        self.full_path =    Path(in_path_pp)
        self.name =         self.full_path.name
        self.is_absolute =  self.full_path.is_absolute()
        self.is_relative =  not self.is_absolute

        if set_attributes:
            self.refresh_stats(timeout=timeout, ntries=ntries)
        else:
            self.exists = self.is_file = self.is_dir = None


#=====================================================================================
#=====================================================================================
#  r e f r e s h _ s t a t s
#=====================================================================================
#=====================================================================================

    def refresh_stats(self, timeout=1.0, ntries=1):
        """
## refresh_stats (timeout=1.0, ntries=1) - Update the instance .exists, .is_dir, and .is_file booleans attributes

***mungePath() class member function***

The instance attributes `.exists`, `.is_dir`, and `.is_file` may be set 
at the time the mungePath instance is created by setting `set_attributes=True` (the default is False). 
These attributes are not updated automatically as changes happen on the filesystem. 
Call `refresh_stats()` as needed, or directly access the pathlib methods (or access through
`run_with_timeout()`), eg `my_mungepath_inst.full_path.exists()`.  Also see `check_path_exists()`.

NOTE:  `refresh_stats()` utilizes `run_with_timeout()` for each of the three attribute settings.
refresh_stat's default timeout=1 second and ntries=1 can result in up to a 3 second _hang_ if there 
are access issues.


### Args
`timeout` (int or float, default 1.0 second)
- The timeout value passed to each call to `run_with_timeout()`

`ntries` (int, default 1)
- Number of tries value passed to each call to `run_with_timeout()`


### Returns
- The instance handle is returned so that refresh_stats() may be used in-line.
- Sets each attribute to valid True/False value if possible, else sets the attribute to False on timeout error.
- Sets each attribute to False on any exception.
        """
        self.exists = check_path_exists(self.full_path, timeout=timeout, ntries=ntries)

        try:
            self.is_dir = run_with_timeout (self.full_path.is_dir, rwt_timeout=timeout, rwt_ntries=ntries)
        except Exception as e:
            logging.debug(f"Exception - {type(e).__name__}: {e}")
            self.is_dir = False

        try:
            self.is_file = run_with_timeout (self.full_path.is_file, rwt_timeout=timeout, rwt_ntries=ntries)
        except Exception as e:
            logging.debug(f"Exception - {type(e).__name__}: {e}")
            self.is_file = False

        return self


#=====================================================================================
#=====================================================================================
#  _ _ r e p r _ _
#=====================================================================================
#=====================================================================================

    def __repr__(self):
        stats = ""
        stats +=  f".full_path    :  {self.full_path}\n"
        stats +=  f".parent       :  {self.parent}\n"
        stats +=  f".name         :  {self.name}\n"
        stats +=  f".is_absolute  :  {self.is_absolute}\n"
        stats +=  f".is_relative  :  {self.is_relative}\n"
        stats +=  f".exists       :  {self.exists}\n"
        stats +=  f".is_dir       :  {self.is_dir}\n"
        stats +=  f".is_file      :  {self.is_file}\n"
        return stats


#=====================================================================================
#=====================================================================================
#  c h e c k _ p a t h _ e x i s t s
#=====================================================================================
#=====================================================================================

def check_path_exists(inpath, timeout=1.0, ntries=1):
    """
## check_path_exists (inpath, timeout=1.0, ntries=1) - With enforced timeout (no hang)

pathlib.Path.exists() tends to hang for an extended time when there are network access issues.
check_path_exists() wraps `pathlib.Path.exists()` with a call to run_with_timeout().


### Args
`inpath` (Path or str)
- Path to a file or directory

`timeout` (int or float, default 1.0 second)
- timeout value used in `run_with_timeout()` call

`ntries` (int, default 1)
- Number of tries value passed to `run_with_timeout()`


### Returns
- True if the path exists
- False if the path does not exist or the timeout is reached (or any other exception)
    """

    _path = Path(inpath)

    try:
        return run_with_timeout (_path.exists, rwt_timeout=timeout, rwt_ntries=ntries)
    except Exception as e:
        logging.debug(f"Exception - {type(e).__name__}: {e}")
        return False



#!/usr/bin/env python3
"""Inter-process lock mechanism using posix_ipc

Only works on Linux
"""

#==========================================================
#
#  Chris Nelson, Copyright 2024-2025
#
#==========================================================

import signal
import sys
import posix_ipc
import mmap
import os
import datetime
from .core import logging, set_toolname, setuplogging

import importlib.metadata
__version__ = importlib.metadata.version(__package__ or __name__)


# Logging events within this module are at the DEBUG level.  With this module's child logger set to
# a minimum of WARNING level by default, then logging from this module is effectively disabled.  To enable
# logging from this module add this within your tool script code:
#       logging.getLogger('cjnfuncs.resourcelock').setLevel(logging.DEBUG)
resourcelock_logger = logging.getLogger('cjnfuncs.resourcelock')
resourcelock_logger.setLevel(logging.WARNING)

# Logging from the .is_locked method is enabled by:
#       logging.getLogger('cjnfuncs.resourcelock_islocked').setLevel(logging.DEBUG)
resourcelock_logger_islocked = logging.getLogger('cjnfuncs.resourcelock_islocked')
resourcelock_logger_islocked.setLevel(logging.WARNING)


#=====================================================================================
#=====================================================================================
#  r e s o u r c e _ l o c k
#=====================================================================================
#=====================================================================================

class resource_lock():
    """
## Class resource_lock (lockname) - Inter-process lock mechanism using posix_ipc

__NOTE:  This module only works on Linux.__

In applications that have independent processes sharing a resource, such as an I2C bus, `resource_lock()`
provides a semaphore communication mechanism between the processes, using the posix-ipc module, 
in order to coordinate access to the shared resource.  By using resource_lock(), ProcessA becomes aware
that the I2C bus is in-use by some other process (ProcessB), and it should wait until that other 
process completes its work, and then acquire the I2C bus lock so that other process(es) are blocked. 

- Resource locks are on the honor system.  Any process can unget a lock, but should not if it didn't get the lock.

- This lock mechanism is just as effective across threads within a process, and between processes.

- As many different/independent locks as needed may be created.

- The first time a lock is created (on the current computer since reboot) the lock info string is set to ''
(accessible via `get_lock_info()`), else it retains the value set by the most recent get_lock() call.

- It is recommended (in order to avoid a minor memory leak) to `close()` the lock in the tool script cleanup code.
Calling `close()` sets the `closed` attribute to True so that any following code within the current tool script
can detect and re-instantiate the lock if needed.

- Semaphores (lock names) and shared memory segments (used for the `lock_info` string) in the posix_ipc module 
must have `/` prefixes.  resource_lock() prepends the `/` if `lockname` doesn't start with a `/`, and hides the `/` prefix.

resource_lock() requires the `posix_ipc` module (installed with cjnfuncs) from PyPI. 
See https://pypi.org/project/posix-ipc/.

NOTE that a crashed process may not have released the lock, resulting in other processes using the lock to hang.
Use the CLI command `resourcelock <lockname> unget` to manually release the lock to un-block other processes.

resource_lock() uses `posix_ipc.Semaphore`, which is a counter mechanism. `get_lock()` 
decrements the counter to 0, indicating a locked state.  `unget_lock()` increments the
counter (non-zero is unlocked). `unget_lock()` wont increment the counter unless the counter is 
currently 0 (indicating locked), so it is ***recommended*** to have (possibly extraneous) `unget_lock()` calls, 
such as in your interrupt-trapped cleanup code.


### Args
`lockname` (str)
- All processes sharing a given resource must use the same lockname.

### Class attributes
`lockname` (str)
- As specified when the resource_lock was instantiated

`I_have_the_lock` (bool)
- True if the current process has set the lock.  Useful for conditionally ungetting the lock in cleanup code.

`closed` (bool)
- False once instantiated and set True if `close()` is called in script cleanup code, so that the lock can 
checked and re-instantiate if needed.
    """

    def __init__ (self, lockname):
        if not lockname.startswith('/'):
            lockname = '/'+lockname         # lockname is required to start with '/'
        self.lockname =         lockname
        self.closed =           False
        self.I_have_the_lock =  False
        self.lock = posix_ipc.Semaphore(self.lockname, flags=posix_ipc.O_CREAT, mode=0o0600, initial_value=1)

        preexisting = False
        try:
            memory = posix_ipc.SharedMemory(self.lockname, flags=0)
            preexisting = True
        except posix_ipc.ExistentialError:
            memory = posix_ipc.SharedMemory(self.lockname, flags=posix_ipc.O_CREAT, mode=0o0600, size=4096)

        self.mapfile = mmap.mmap(memory.fd, memory.size)
        os.close(memory.fd)
        if not preexisting:
            self._set_lock_info('')


#=====================================================================================
#=====================================================================================
#  g e t _ l o c k
#=====================================================================================
#=====================================================================================

    def get_lock(self, timeout=1.0, same_process_ok=False, lock_info=''):
        """
## get_lock (timeout=1.0, same_process_ok=False, lock_info='') - Request the resource lock

***resource_lock() class member function***

Attempt to acquire/get the lock while waiting up to `timeout` time.  

By default, get_lock() waits for the lock if it is currently set, whether the lock was set by this
or another script/job/process.

By setting `same_process_ok=True`, then if the lock was previously acquired by this same script/process
then get_lock() immediately returns True.  This allows the script code to not have to track state to 
decide if the lock has previously been acquired before calling get_lock() again, leading to cleaner code.

### Args
`timeout` (int or float, or None, default 1.0 second)
- The max time, in seconds, to wait to acquire the lock
- None is no timeout - wait forever (Hang forever.  Unwise.)

`same_process_ok` (bool, default False)
- If True, then if the current process currently has the lock then get_lock() immediately returns True.
- If False, then if the lock is currently set by the same process or another process then get_lock() blocks
with timeout.

`lock_info` (str, default '')
- Optional debugging info string for indicating when and by whom the lock was set.  Logged at the debug level.
- The datetime is prepended to lock_info.
- A useful lock_info string format might be `<module_name>.<function_name> <get_lock_call_instance_number>`, eg, 
`tempmon.measure_loop #3`.
- This string remains in place after an unget() call (`is_locked() == False`) for lock history purposes while debugging.

### Returns
- True:  Lock successfully acquired, timeout time not exceeded
- False: Lock request failed, timed out
        """
        if same_process_ok  and  self.I_have_the_lock == True:
            resourcelock_logger.debug (f"<{self.lockname[1:]}> lock already acquired   - Prior grant     <{self.get_lock_info()}>")
            return True

        try:
            self.lock.acquire(timeout)
            lock_text = f"{datetime.datetime.now()} - {lock_info}"
            self._set_lock_info(lock_text)
            self.I_have_the_lock = True
            resourcelock_logger.debug (f"<{self.lockname[1:]}> lock request successful - Granted         <{lock_text}>")
            return True
        except posix_ipc.BusyError:
            resourcelock_logger.debug (f"<{self.lockname[1:]}> lock request timed out  - Current owner   <{self.get_lock_info()}>")
            return False


#=====================================================================================
#=====================================================================================
#  u n g e t _ l o c k
#=====================================================================================
#=====================================================================================

    def unget_lock(self, force=False, where_called=''):
        """
## unget_lock (force=False, where_called='') - Release the resource lock

***resource_lock() class member function***

If the lock was acquired by the current process then release the lock.
- If the lock is not currently set then the `unget_lock()` call is discarded, leaving the lock
in the same unset state.
- If the lock is currently set but _not_ acquired by this process then don't release the lock,
unless `force=True`.

### Arg
`force` (bool, default False)
- Release the lock regardless of whether or not this process acquired it.
- Useful for forced cleanup, for example, by the CLI interface.
- Dangerous if another process had acquired the lock.  Be careful.

`where_called` (str, default '')
- Debugging aid string for indicating what code released the lock.  Logged at the debug level.
Not stored anywhere, nor available to a later call.

### Returns
- True:  Lock successfully released
- False: Lock not currently set (redundant unget_lock() call), or lock was not acquired by the current process
        """
        if self.lock.value == 0:
            if self.I_have_the_lock:
                self.lock.release()
                self.I_have_the_lock = False
                resourcelock_logger.debug (f"<{self.lockname[1:]}> lock released  <{where_called}>")
                return True
            else:
                if force:
                    self.lock.release()
                    resourcelock_logger.debug (f"<{self.lockname[1:]}> lock force released  <{where_called}>")
                    return True
                else:
                    resourcelock_logger.debug (f"<{self.lockname[1:]}> lock unget request ignored - lock not owned by current process  <{where_called}>")
                    return False
        else:
            resourcelock_logger.debug (f"<{self.lockname[1:]}> Extraneous lock unget request ignored  <{where_called}>")
            return False


#=====================================================================================
#=====================================================================================
#  i s _ l o c k e d
#=====================================================================================
#=====================================================================================

    def is_locked(self):
        """
## is_locked () - Returns the current state of the lock

***resource_lock() class member function***

### Returns
- True if currently locked, else False
        """
        locked = True  if self.lock.value == 0  else False
        resourcelock_logger_islocked.debug (f"<{self.lockname[1:]}> is currently locked?  <{locked}>  Prior info  <{self.get_lock_info()}>")
        return locked


#=====================================================================================
#=====================================================================================
#  l o c k _ v a l u e
#=====================================================================================
#=====================================================================================

    def lock_value(self):
        """
## lock_value () - Returns the lock semaphore count

***resource_lock() class member function***

### Returns
- Current value of the semaphore count - should be 0 (locked) or 1 (unlocked)
        """
        _value = self.lock.value
        resourcelock_logger.debug (f"<{self.lockname[1:]}> semaphore = {_value}")
        return _value


#=====================================================================================
#=====================================================================================
#  c l o s e
#=====================================================================================
#=====================================================================================

    def close(self):
        """
## close () - Release this process' access to the semaphore and the memory-mapped shared memory segment

***resource_lock() class member function***

### Returns
- None
        """
        self.lock.close()
        self.mapfile.close()
        self.closed = True
        resourcelock_logger.debug (f"<{self.lockname[1:]}> semaphore closed")


#=====================================================================================
#=====================================================================================
#  g e t _ l o c k _ i n f o
#=====================================================================================
#=====================================================================================

    def get_lock_info(self):
        """
## get_lock_info () - Returns the lock_info string from previous get_lock call

***resource_lock() class member function***

### Returns
- lock_info string
        """
        self.mapfile.seek(0)
        s = []
        c = self.mapfile.read_byte()
        while c != 0:    # NULL_CHAR
            s.append(c)
            c = self.mapfile.read_byte()

        s = ''.join([chr(c) for c in s])

        return s


#=====================================================================================
#=====================================================================================
#  _ s e t _ l o c k _ i n f o   (private function)
#=====================================================================================
#=====================================================================================

    def _set_lock_info(self, desc):
        # While the shared memory segment and memory mapped block are 4k bytes log, the actual 
        # lock_info description is terminated by a null character (0x00)
        self.mapfile.seek(0)
        desc += '\0'
        self.mapfile.write(desc.encode())
       


#=====================================================================================
#=====================================================================================
#  c l i
#=====================================================================================
#=====================================================================================

def int_handler(sig, frame):
    resourcelock_logger.warning(f"Signal {sig} received")
    sys.exit(0)

signal.signal(signal.SIGINT,  int_handler)      # Ctrl-C
signal.signal(signal.SIGTERM, int_handler)      # kill


def cli():
    docplus = """
    Commands:
        get:    Get/set the lock named LockName.  '-a' specifies a automatic timed unget (only applied if the get was successful).
        unget:  Force-release LockName.
        state:  Print the current state of LockName.
        trace:  Continuously print the state of LockName.  '-u' specifies update interval.  Ctrl-C to exit.
    """
    import argparse
    from time import sleep

    set_toolname ('resourcelock_cli')
    setuplogging()
    logging.getLogger('cjnfuncs.resourcelock').setLevel(logging.DEBUG)
    logging.getLogger('cjnfuncs.resourcelock_islocked').setLevel(logging.DEBUG)

    GET_TIMEOUT = 0.5
    TRACE_INTERVAL = 0.5
    

    parser = argparse.ArgumentParser(description=__doc__ + __version__+docplus, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('LockName',
                        help="Name of the system-wide lock to access")
    parser.add_argument('Cmd', choices=['get', 'unget', 'state', 'trace'],
                        help="Command choices")
    parser.add_argument('-t', '--get-timeout', type=float, default=GET_TIMEOUT,
                        help=f"Timeout value for a get call (default {GET_TIMEOUT} sec, -1 for no timeout)")
    parser.add_argument('-m', '--message', default='cli',
                        help=f"Lock get/unget debug message text (default 'cli')")
    parser.add_argument('-a', '--auto-unget', type=float,
                        help="After a successful get, unget the lock in (float) sec")
    parser.add_argument('-u', '--update', type=float, default=TRACE_INTERVAL,
                        help=f"Trace update interval (default {TRACE_INTERVAL} sec)")
    args = parser.parse_args()

    lock = resource_lock(args.LockName)


    if args.Cmd == "get":
        _timeout = args.get_timeout
        if _timeout == -1:
            _timeout = None
        get_status = lock.get_lock(timeout=_timeout, lock_info=args.message)
        if get_status and args.auto_unget:
                print (f"Release lock after <{args.auto_unget}> sec delay")
                sleep(args.auto_unget)
                lock.unget_lock(where_called=f'{args.message} - auto unget')

    elif args.Cmd == "unget":
        lock.unget_lock(force=True, where_called=args.message)

    elif args.Cmd == "state":
        lock.is_locked()

    elif args.Cmd == "trace":
        while True:
            lock.is_locked()
            sleep (args.update)

    else:
        print ("How did we get here?")



#!/usr/bin/env python3
"""cjnfuncs.configman - Configuration files made easy and useful
"""

#==========================================================
#
#  Chris Nelson, 2018-2025
#
#==========================================================

import re
import ast

from .core      import setuplogging, logging, ConfigError #, set_logging_level, restore_logging_level, pop_logging_level_stack, get_logging_level_stack
from .mungePath import mungePath, check_path_exists
from .rwt       import run_with_timeout
import cjnfuncs.core as core

# Configs / Constants
RWT_NTRIES =    3
RWT_TIMEOUT =   2.0

# Logging events within this module are at the INFO and DEBUG levels.  With this module's child logger set to
# a minimum of WARNING level by default, then logging from this module is effectively disabled.  To enable
# logging from this module add this within your tool script code:
#       logging.getLogger('cjnfuncs.configman').setLevel(logging.DEBUG)
configman_logger = logging.getLogger('cjnfuncs.configman')
configman_logger.setLevel(logging.WARNING)


#=====================================================================================
#=====================================================================================
#  C l a s s   c o n f i g _ i t e m
#=====================================================================================
#=====================================================================================

initial_logging_setup_done = False   # Global since more than one config can be loaded


class config_item():
    """
## Class config_item (config_file=None, remap_logdirbase=True, force_str=False, secondary_config=False, safe_mode=False) - Create a configuration instance
The config_item() class provides handling of one or more config file instances.  Class methods include:
 - Config file loading and reloading - `loadconfig()`
 - Loading config data from strings and dictionaries - `read_string()`, `read_dict()`
 - Getting values from the loaded config, with defaults and fallback - `getcfg()`
 - Programmatically modifying the config file content - `modify_configfile()`
 - Getting instance status - `__repr__()`, `section()`, `dump()`

See the loadconfig() documentation for details on config file syntax and rules.


### Instantiation args
`config_file` (Path or str, default None)
- Path to the configuration file, relative to the `core.tool.config_dir` directory, or an absolute path.
- `None` may be used if the config will be loaded programmatically via `read_string()` or `read_dict()`.

`remap_logdirbase` (bool, default True)
- If `remap_logdirbase=True` and the tool script is running in _user_ mode (not site mode) 
then the `core.tool.log_dir_base` will be set to `core.tool.config_dir`.

`force_str` (bool, default False)
- Causes all params to be loaded as type `str`, overriding the default type identification.

`secondary_config` (bool, default False)
- Set to `True` when loading additional config files.  Disables logging setup related changes.
- The primary config file should be loaded first before any secondary_config loads, so that logging 
is properly set up.

`safe_mode` (bool, default False)
- If `safe_mode=True` then timeouts are enforced when checking for config files.
- If `safe_mode=False` then checks for the existence of config files runs the risk of application hang.
- See Behavior NOTE below.

### Useful class attributes
The current values of all public class attributes may be printed using `print(my_config)`.

`.cfg` (dict)
- Holds all loaded params.  May be access directly.  Sections are stored as sub-dictionaries of .cfg.
- The contents of the .cfg dictionary may be printed using `print(my_config.dump())`

`.defaults` (dict)
- Default params are stored here.

`.sections_list` (list)
- A list of string names for all defined sections.

`.config_file` (str, or None)
- The `config_file` as passed in at instantiation

`.config_full_path` (Path or None)
- The full expanduser/expandvars path to the config file, relative to core.tool.config_dir if the
instantiation `config_file` is a relative path (uses mungePath)

`.config_dir` (Path or None)
- The directory above  `.config_full_path`


### Returns
- Handle to the `config_item()` instance
- Raises a `ConfigError` if the specified config file is not found


### Behaviors and rules
1. More than one `config_item()` may be created and loaded.  This allows for configuration data to be partitioned 
as desired.  Each defined config is loaded to its own instance-specific `cfg` dictionary. Only one config_item()
instance should be considered the primary, while other instances should be tagged with `secondary_config=True`. 
Logging setups are controlled only by the primary config instance.
Also see the loadconfig() `import` feature.

1. NOTE:  On Linux the cost of `safe_mode=True` when searching for a quickly found config file is negligible 
(Linux uses process fork), but on Windows (uses process spawn) the minimum execution time for a file 
existence check can be 1 to 2 seconds.  When `safe_mode=False` and the config file is on a network share 
drive, the file existence check can cause a indefinite blocking hang of your tool script if the network 
goes down (true for both Linux and Windows).  So, use `safe_mode=False` if your config files are on a local 
drive, and use `safe_mode=True` if your config files are on a network drive.  Recognize that on Windows 
this will perhaps cause significant delays.

1. Initially in _user_ mode, after the `set_toolname()` call, `core.tool.log_dir_base` 
(the log directory) is set to the `core.tool.user_data_dir`.
Once `config_item()` is called the `core.tool.log_dir_base` is _remapped_ to 
`core.tool.config_dir`.  This is the author's style preference (centralize primary files, and 
reduce spreading files around the file system).
To disable this remap, in the `config_item()` call set `remap_logdirbase=False`.
This remapping is not done in site mode.

1. A different log base directory may be set by user code by setting `core.tool.log_dir_base` to a different 
path after the `set_toolname()` call and before the `config_item()` call, for example 
`core.tool.log_dir_base = "/var/log"` may be desirable in site mode.

1. A different config directory may be set by user code by setting `core.tool.config_dir` to a different 
path after the `set_toolname()` call and before the `config_item()` call, for example 
`core.tool.config_dir = core.tool.main_dir` sets the config dir to the same as the tool script's 
directory.  With `remap_logdirbase=True`, the log dir will also be set to the tool script's directory.

1. Details of the configuration instance may be printed, eg, `print(my_config)`.

1. Logging of the operations within loadconfig() and (other configman methods) may be controlled with setting 
the named child logger `cjnfuncs.configman` to INFO or DEBUG level (the default logging level is WARNING which
produces no logging events from this module), eg:

        logging.getLogger('cjnfuncs.configman').setLevel(logging.INFO)
    """

    def __init__(self, config_file=None, remap_logdirbase=True, force_str=False, secondary_config=False, safe_mode=False):
        global tool

        self.force_str =            force_str
        self.secondary_config =     secondary_config
        self.safe_mode =            safe_mode
        self.cfg =                  {}
        self.current_section_name = ''
        self.sections_list =        []
        self.defaults =             {}

        if config_file == None:
            self.config_file =      None
            self.config_dir =       None
            self.config_full_path = None
            self.config_timestamp = 0
            self.config_content =   ''        # Used by modify_configfile()
        else:
            config_mp = mungePath(config_file, core.tool.config_dir)
            found_it = False
            if self.safe_mode:
                if run_with_timeout(config_mp.full_path.is_file, rwt_ntries=RWT_NTRIES, rwt_timeout=RWT_TIMEOUT):
                    found_it = True
            else:
                if config_mp.full_path.is_file():
                    found_it = True
            if found_it:
                self.config_file        = config_mp.name
                self.config_dir         = config_mp.parent
                self.config_full_path   = config_mp.full_path
                self.config_timestamp   = 0
                self.config_content     = ''    # Used by modify_configfile()
            else:
                raise ConfigError (f"Config file <{config_file}> not found.")

        if remap_logdirbase  and  core.tool.log_dir_base == core.tool.user_data_dir:
            core.tool.log_dir_base = core.tool.config_dir


    def _add_key(self, key, value, section_name=''):
        if section_name == '':
            self.cfg[key] = value
        elif section_name == 'DEFAULT':
            self.defaults[key] = value
        else:
            self.cfg[section_name][key] = value


#=====================================================================================
#=====================================================================================
#  l o a d c o n f i g
#=====================================================================================
#=====================================================================================

    def loadconfig(self,
            call_logfile        = None,
            call_logfile_wins   = False,
            flush_on_reload     = False,
            force_flush_reload  = False,
            isimport            = False,
            tolerate_missing    = False,
            prereload_callback  = None):
        r"""
## loadconfig () - Load a configuration file into the cfg dictionary
```
loadconfig(
    call_logfile        = None,
    call_logfile_wins   = False,
    flush_on_reload     = False,
    force_flush_reload  = False,
    isimport            = False,
    tolerate_missing    = False,
    prereload_callback  = None)        
```
***config_item() class member function***

`Param = value` lines in the config_item()'s file are loaded to the instance-specific `cfg` dictionary, 
and can be accessed via `<config_item>.getcfg()`.  The _value_ is referred to as the _value_portion_ in 
this documentation.

`loadconfig()` initializes the root logger for logging either to 1) the `LogFile` specified in
the loaded config file, 2) the `call_logfile` in the `loadconfig()` call, or 3) the console.
`loadconfig()` supports dynamic reloading of config files, partitioning of config data via the `import`
feature, and intermittent loss of access to the config file.
    

### Args
`call_logfile` (Path or str, default None)
- If `call_logfile` is passed on the loadconfig() call, and `call_logfile_wins=True`, then any `LogFile`
specified in the config file is overridden.  This feature allows for interactive usage modes where
logging is directed to the console (with `call_logfile=None`) or an alternate file.
- An absolute path or relative to the `core.tool.log_dir_base` directory

`call_logfile_wins` (bool, default False)
- If True, the `call_logfile` overrides any `LogFile` defined in the config file

`flush_on_reload` (bool, default False)
- If the config file will be reloaded (due to a changed timestamp) then clean out the 
`cfg` dictionary first.  See Returns, below.

`force_flush_reload` (bool, default False)
- Forces the `cfg` dictionary to be cleaned out and the config file to be reloaded, 
regardless of whether the config file timestamp has changed

`isimport` (bool, default False)
- Internally set True when handling imports.  Not used by tool script calls.

`tolerate_missing` (bool, default False)
- Used in a tool script service loop, return `-1` rather than raising `ConfigError` if the config file is inaccessible

`prereload_callback` (function, default None)
- Allows user services to be managed (paused/terminated) before the config is reloaded and logging is reset.

### Returns
- `1` if the config files WAS reloaded
- `0` if the config file was NOT reloaded
- If the config file cannot be accessed
  - If tolerate_missing == False (default), then raises `ConfigError`
  - If tolerate_missing == True, then returns `-1`
- A ConfigError is raised if there are parsing issues
- A ConfigError is also raised if an imported config file cannot be loaded (non-existent)


### Behaviors and rules
1. See `getcfg()`, below, for accessing loaded config data. The class instance-specific `cfg` dictionary may be
  directly accessed as well.

1. The format of a config file is param=value pairs.
   - Separating the param and value_portion may be whitespace, `=` or `:` (multiples allowed).
   - Param names can contain all valid characters, except the separators or `#`, and cannot start with `[`.

1. Sections and a DEFAULT section are supported.  Section name are enclosed in `[ ]`.
   - Leading and trailing whitespace is trimmed off of the section name, and embedded whitespace is retained.
    EG: `[  hello my name  is  Fred  ]` becomes section name `'hello my name  is  Fred'`.
   - Section names can contain most all characters, except `]`.

1. **Native int, float, bool, list, tuple, dict, str support** - Bool true/false is case insensitive. A str
  type is stored in the `cfg` dictionary if none of the other types can be resolved for a given value_portion.
  Automatic typing avoids most explicit type casting clutter in the tool script. Be careful to error trap
  for type errors (eg, expecting a float but user input error resulted in a str). Also see the 
  getcfg() `types=[]` arg for basic type enforcement.

1. **Quoted strings** - If a value_portion cannot be resolved to a Python native type then it is loaded as a str,
  eg `My_name = George` loads George as a str.  A value_portion may be forced to be loaded as a str by using 
  quotes, eg `Some_number_as_str = "2.54"` forces the value_portion to be loaded as a str rather than a float. Supported
  quote types:  `"..."`, `'...'`, (triple-double quotes), and `'''...'''`. `My_name = George`, 
  `My_name : "George"`, `My_name '''George'''`, etc., are identical when loaded.
  Quoted strings may contain all valid characters, including '#' which normally starts a comment.  

1. **Multi-line values** - A param's value_portion may be specified over multiple lines for readability by placing 
  the `\` line continuation character as the last non-whitespace character on the line before any comment.
  The parser strips comments and leading/trailing whitespace, then concatenates the multi-line value_portion segments 
  into a single line (single space separated) in the loaded config.  Comments may be placed on each line.
  NOTE: For a multi-line param that will be loaded as a str, avoid using quotes as results may be strange.

1. **Logging setup** - `loadconfig()` calls `cjnfuncs.core.setuplogging()`.  The `logging` handle is available for
  import by other modules (`from cjnfuncs.core import logging`).  By default, logging will go to the
  console (stdout) filtered at the WARNING/30 level. Don't call `setuplogging()` directly if using loadconfig().

1. **Logging level control** - Optional `LogLevel` in the primary config file will set the root logging level after
  the config file has been loaded.  If LogLevel is not specified in the primary config file, then 
  the root logging level is left unchanged (the Python default logging level is 30/WARNING).
  The tool script code may also manually/explicitly set the root logging level _after the initial `loadconifig()` call_
  and this value will be retained over later calls to loadconfig, thus allowing for a command line `--verbose`
  switch feature.  Note that logging done _within_ loadconfig() uses the `cjnfuncs.configmap` child/named logger.
  `logging.getLogger('cjnfuncs.configman').setLevel(logging.INFO)` (or DEBUG) enables diagnostic logging from
  loadconfig.

1. **Log file options** - Where to log has two separate fields:  `call_logifle` in the call to loadconfig(), and 
  `LogFile` in the loaded primary config file, with `call_logfile_wins` selecting which is used.  This mechanism allows for
  a command line `--log-file` switch to override a _default_ log file defined in the config file.  If the selected 
  logging location is `None` then output goes to the console (stdout).

    call_logfile_wins | call_logfile | Config LogFile | Results
    --|--|--|--
    False (default) | ignored | None (default) | Console
    False (default) | ignored | file_path | To the config LogFile
    True | None (default) | ignored | Console
    True | file_path | ignored | To the call_logfile

1. **Logging format** - cjnfuncs has default format strings for console and file logging.
  These defaults may be overridden by defining `ConsoleLogFormat` and/or `FileLogFormat`
  in the config file.

1. **Import nested config files** - loadconfig() supports `Import` (case insensitive). The imported file path
is relative to the `core.tool.config_dir`, if not an absolute path.
The specified file is imported as if the params were in the main config file.  Nested imports are allowed. 
Sections are not allowed within an imported file - only in the main/top-level config file.
A prime usage of `import` is to place email server credentials in your home directory with user-only readability,
then import them in the tool script config file as such: `import ~/creds_SMTP`.  

1. **Config reload if changed, `flush_on_reload`, and `force_flush_reload`** - loadconfig() may be called 
periodically by the tool script, such as in a service loop.
If the config file timestamp is unchanged then loadconfig() immediately returns `0`. 
If the timestamp has changed then the config file will be reloaded and `1` is returned to indicate to 
the tool script to do any post-config-load operations. 
   - If `flush_on_reload=True` (default False) then the instance-specific `cfg` dictionary 
  will be cleaned/purged before the config file is reloaded. If `flush_on_reload=False` then the config
  file will be reloaded on top of the existing `cfg` dictionary contents (if a param was 
  deleted in the config
  file it will still exist in `cfg` after the reload). [lanmonitor](https://github.com/cjnaz/lanmonitor) uses the
  `flush_on_reload=True` feature.
   - `force_flush_reload=True` (default False) forces both a clear/flush of the `cfg` dictionary and then a fresh
  reload of the config file. 
   - **Note** that if using threading then a thread should be paused while the config file 
  is being reloaded with `flush_on_reload=True` or `force_flush_reload=True` since the params will disappear briefly.
  Use the `prereload_callback` mechanism to manage any code dependencies before the cfg dictionary is purged.
   - Changes to imported files are not tracked for changes.

1. **Tolerating intermittent config file access** - When implementing a service loop, if `tolerate_missing=True` 
(default False) then loadconfig() will return `-1` if the config file cannot be accessed, informing the 
tool script of the problem for appropriate handling (typically logging the event then ignoring the problem for
the current iteration). If `tolerate_missing=False` then loadconfig() will raise a ConfigError if the config file 
cannot be accessed.
        """

        global initial_logging_setup_done

        if not initial_logging_setup_done:
            # Initial logging will go to the console if no call_logfile is specified (and call_logfile_wins) on the initial loadconfig call.
            console_lf = self.getcfg('ConsoleLogFormat', None)
            file_lf = self.getcfg('FileLogFormat', None)
            setuplogging (call_logfile=call_logfile, call_logfile_wins=call_logfile_wins, ConsoleLogFormat=console_lf, FileLogFormat=file_lf)
            initial_logging_setup_done = True

        config = self.config_full_path

        # Operations only on top-level config file
        is_top_level = not isimport
        if is_top_level:

            if force_flush_reload:
                configman_logger.info(f"Config  <{self.config_file}>  force flushed (force_flush_reload)")
                self.clear()
                self.config_timestamp = 0                       # Force reload of the config file

            # Check if config file is available and has changed
            _exists = False
            if self.safe_mode:
                if check_path_exists(config, ntries=RWT_NTRIES, timeout=RWT_TIMEOUT):
                    _exists = True
            else:
                if config.exists():
                    _exists = True

            if not _exists:
                if tolerate_missing:
                    configman_logger.info (f"Config  <{self.config_file}>  file not currently accessible.  Skipping (re)load.")
                    return -1                                   # -1 indicates that the config file cannot currently be found
                else:
                    raise ConfigError (f"Could not find  <{self.config_file}>")

            current_timestamp = int(self.config_full_path.stat().st_mtime)  # force integer-second resolution
            if self.config_timestamp == current_timestamp:
                return 0                                        # 0 indicates that the config file was NOT (re)loaded

            # It's an initial load call, or config file has changed, or force_flush_reload...  Do (re)load
            self.config_timestamp = current_timestamp
            if prereload_callback:
                configman_logger.debug("Pre-reload callback user function called")
                prereload_callback()
            configman_logger.info (f"Config  <{self.config_file}>  file timestamp: {current_timestamp}")

            if flush_on_reload:
                configman_logger.info (f"Config  <{self.config_file}>  flushed due to changed file (flush_on_reload)")
                self.clear()


        # Load the config
            # If there is an import then read_string() recursively calls loadconfig().  Imported configs can in-turn import configs.
            # An exception raised within read_string() while processing an imported config lands at this except clause, and is re-raised
            # to be handled at the next higher level, until we reach the top-level loadconfig() call, and the exception is passed to the 
            # tool script code.
        configman_logger.info (f"Loading  <{config}>")
        try:
            string_blob = config.read_text()
            self.read_string (string_blob, isimport=isimport)
        except:
            raise


        # Operations only for finishing a top-level call on the primary config
        if is_top_level  and  not self.secondary_config:

            # Re-setup the root logger, adjusting for config LogFile and logging formats
            console_lf = self.getcfg('ConsoleLogFormat', None)
            file_lf =    self.getcfg('FileLogFormat', None)
            setuplogging(config_logfile=self.getcfg('LogFile', None), call_logfile=call_logfile, call_logfile_wins=call_logfile_wins, ConsoleLogFormat=console_lf, FileLogFormat=file_lf)

            # Apply SMTP module control flags based on config params
            if 'SMTP' in self.sections_list:
                if self.getcfg('DontEmail', False, section='SMTP'):
                    configman_logger.info ("DontEmail is set - Emails and Notifications will NOT be sent")
                elif self.getcfg('DontNotif', False, section='SMTP'):
                    configman_logger.info ("DontNotif is set - Notifications will NOT be sent")

            # Apply LogLevel, if defined in the config
            config_loglevel = self.getcfg('LogLevel', None)
            if config_loglevel is not None:
                try:
                    config_loglevel = int(config_loglevel)      # Handles force_str=True
                except:
                    raise ConfigError (f"Config file <LogLevel> must be integer value (found <{config_loglevel}>)") from None
                configman_logger.info (f"Logging level set to config LogLevel <{config_loglevel}>")
                logging.getLogger().setLevel(config_loglevel)

        self.current_section_name = ''
        return 1                                                # 1 indicates that the config file was (re)loaded


#=====================================================================================
#=====================================================================================
#  r e a d _ s t r i n g
#=====================================================================================
#=====================================================================================

    def read_string(self, str_blob, isimport=False):
        """
## read_string (str_blob, isimport=False) - Load content of a string into the cfg dictionary

***config_item() class member function***

read_string() does the actual work of loading lines of config data into the cfg dictionary. 
Loaded content is added to and/or modifies any previously loaded content.

Note that loadconfig() calls read_string() for the actual loading of config data. loadconfig()
handles the other loading features such as LogLevel, LogFile, logging formatting,
flush_on_reload, force_flush_reload, and tolerate_missing.


### Args
`str_blob` (str)
- String containing the lines of config data

`isimport` (bool, default False)
- Internally set True when handling imports.  Not used by tool script calls.


### Returns
- A ConfigError is raised if there are parsing issues
- A ConfigError is also raised if an imported config file cannot accessed
        """
        continuation_line = False
        split_line_re =     re.compile(r'([^\s=:#]+)[\s=:]*(.*)')       # Identify param - value (with possible comment)
        xx = fr"""
            ('''|\""")                      # Match opening triple quotes
            (?:\\.|(?!\1).)*?\1             # Match content inside triple quotes until matching closing
            |                               # OR
            (["'])(?:\\.|(?!\2).)*?\2       # Match single or double quotes with matching pairs
            |                               # OR
            (\#)                            # Match unquoted # (group 3)
        """
        find_comment_re =   re.compile(xx, re.VERBOSE)                  # Find unquoted '#'
        section_name_re =   re.compile(r'\[([^\]]*)\]')                 # Get section name


        for line in str_blob.split('\n'):

            if line.strip().startswith('['):                            # Section
                section_name = None
                out = section_name_re.match(line.strip())
                if out:
                    section_name = out.group(1).strip()
                    if section_name != ''  and  section_name not in self.sections_list  and  section_name != 'DEFAULT':
                        self.cfg[section_name] = {}
                        self.sections_list.append(section_name)
                if section_name is None:
                    raise ConfigError (f"Malformed section line <{line}>")
                else:
                    if isimport:
                        raise ConfigError ("Section within imported file is not supported.")
                    self.current_section_name = section_name
                continue

            if not continuation_line:
                param_name = value_portion = ''
                out = split_line_re.match(line.strip())
                if out:
                    param_name    = out.group(1)
                    value_portion = out.group(2)
            else:
                value_portion += ' ' + line.strip()
                continuation_line = False

            if value_portion != '':
                # Remove after first '#' outside of quotes
                for match in find_comment_re.finditer(value_portion):
                    if match.group(3):
                        value_portion = value_portion[:match.start(3)].strip()
                        break

            if value_portion.endswith('\\'):                            # Continuation line
                value_portion = value_portion[:-1].strip()
                continuation_line = True
                continue

            if param_name != '':
                if param_name.lower().startswith('import'):             # import line
                    target = mungePath(value_portion, self.config_dir).full_path
                    imported_config = config_item(target, secondary_config=True)
                    imported_config.loadconfig(isimport=True)           # May raise exception, caught by higher level
                    for key in imported_config.cfg:
                        if self.current_section_name == '':
                            self.cfg[key] = imported_config.cfg[key]
                        elif self.current_section_name == 'DEFAULT':
                            self.defaults[key] = imported_config.cfg[key]
                        else:
                            self.cfg[self.current_section_name][key] = imported_config.cfg[key]
                else:                                                   # param - value line
                    if value_portion.lower() == 'true':
                        value_portion = 'True'
                    if value_portion.lower() == 'false':
                        value_portion = 'False'
                    if value_portion == '':
                        value_portion = 'True'
                    if not self.force_str:
                        try:
                            value_portion = ast.literal_eval(value_portion)
                        except:
                            pass                                        # default to str
                    self._add_key(param_name, value_portion, self.current_section_name)
                    configman_logger.debug (f"Loaded {param_name} = <{value_portion}>  ({type(value_portion)})")


#=====================================================================================
#=====================================================================================
#  read_dict
#=====================================================================================
#=====================================================================================

    def read_dict(self, param_dict, section_name=''):
        """
## read_dict (param_dict, section_name='') - Load the content of a dictionary into the cfg dictionary

***config_item() class member function***

Loaded content is added to and/or modifies any previously loaded content.

### Args
`param_dict` (dict)
- dictionary to be loaded

`section_name` (str, default '' (top level))
- section to load the param_dict into.
- The section will be created if not yet existing.
- Content can only be loaded into one section per call to read_dict().


### Returns
- A ConfigError is raised if there are parsing issues


### Example:
```
    new_config = config_item()      # config need not be associated with a file

    main_contents = {
        'a' : 6,
        'b' : 7.0,
        'c' : [6, 7.0, 42, 'hi']
        }
    new_config.read_dict(main_contents)

    sect_contents = {
        'd' : ('hi', 'there'),
        'e' : {'hi':'Hi!', 'there':'There!'},
        'f' : [6, 7.0, 42, 'hi']
        }
    new_config.read_dict(sect_contents, 'A section')

    def_contents = {
        'g' : 'Hi',
        'h' : True,
        'i' : False
        }
    new_config.read_dict(def_contents, 'DEFAULT')
```
        """

        try:
            if section_name == '':
                for key in param_dict:
                    self.cfg[key] = param_dict[key]
            elif section_name == 'DEFAULT':
                for key in param_dict:
                    self.defaults[key] = param_dict[key]
            else:
                if section_name not in self.sections_list:
                    self.cfg[section_name] = {}
                    self.sections_list.append(section_name)
                for key in param_dict:
                    self.cfg[section_name][key] = param_dict[key]
        except Exception:
            raise ConfigError (f"Failed loading dictionary into cfg around key <{key}>")


#=====================================================================================
#=====================================================================================
#  g e t c f g
#=====================================================================================
#=====================================================================================

    def getcfg(self, param, fallback='_nofallback', types=[], section=''):
        """
## getcfg (param, fallback=None, types=[ ], section='') - Get a param's value from the cfg dictionary

***config_item() class member function***

Returns the value of param from the class instance cfg dictionary.  Equivalent to just referencing `my_config.cfg[]`
but with 1) default & fallback support, 2) type checking, and 3) section support.

The search order for a param is 1) from the specified `section`, 2) from the `DEFAULT` section, and 3) from the 
`fallback` value. If the param is not found in any of these locations then a ConfigError is raised.

Type checking may be performed by listing one or more expected types via the optional `types` arg.
If the loaded param is not one of the expected types then a ConfigError is raised.  This check may be 
useful for basic error checking of param values, eg, making sure the return value is a float and not
a str. (str is the loadconfig() default if the param type cannot be converted to another supported type.)

NOTE: `getcfg()` is almost equivalent to `cfg.get()`, except that `getcfg()` does not default to `None`.
Rather, `getcfg()` raises a ConfigError if the param does not exist and no `fallback` is specified.
This can lead to cleaner tool script code.  Either access method may be used, along with `x = my_config.cfg["param"]`.


### Args
`param` (str)
- String name of param to be fetched from cfg

`fallback` (any, default effectively `None`, technically '_nofallback')
- if provided, is returned if `param` does not exist in cfg
- No type enforcement - the fallback value need not be in the `types` list.

`types` (single or list of as-expected types, default '[]' (any type accepted))
- if provided, a ConfigError is raised if the param's value type is not in the list of expected types
- `types` may be a single type (eg, `types=int`) or a list of types (eg, `types=[int, float]`)
- Supported types: [str, int, float, bool, list, tuple, dict]

`section` (str, default '' (top-level))
- Select the section from which to get the param value.


### Returns
- The param value from 1) from the specified `section` if defined, 2) from the `DEFAULT` section if defined,
  or 3) from the `fallback` value if specified.
- If the param is not found, or the param's type is not in the `types` list, if specified, then a ConfigError is raised.
        """

        _value = None
        if section == '':                           # Top-level case
            if param in self.cfg:
                _value = self.cfg[param]
            elif param in self.defaults:
                _value = self.defaults[param]
        else:
            if section in self.sections_list:       # Section exists case
                if param in self.cfg[section]:
                    _value = self.cfg[section][param]
                elif param in self.defaults:
                    _value = self.defaults[param]
            else:                                   # Section doesn't exist case
                if param in self.defaults:
                    _value = self.defaults[param]

        if _value is None:
            if fallback != '_nofallback':
                return fallback
            else:
                raise ConfigError (f"Param <[{section}] {param}> not in <{self.config_file}> and no default or fallback.")
        else:
            # Optional type checking
            if isinstance(types, type):
                types = [types]

            if types == []:
                    return _value
            else:
                if type(_value) in types:
                    return _value
                else:
                    raise ConfigError (f"Config parameter <[{section}] {param}> value <{_value}> type {type(_value)} not of expected type(s): {types}")


#=====================================================================================
#=====================================================================================
#  m o d i f y _ c o n f i g f i l e
#=====================================================================================
#=====================================================================================

    def modify_configfile (self, param='', value='', remove=False, add_if_not_existing=False, save=False):
        # TODO Comment out / uncomment out a param
        # TODO Add section select parameter for selective changes.  Also all mode
        # TODO Add a section label
        """
## modify_configfile (param='', value='', remove=False, add_if_not_existing=False, save=False) - Make edits to the config file

***config_item() class member function***

Params in the config file may have their values changed, be deleted, or new lines added.
- All added lines are added at the bottom of the file.
- _All instances of the param (in all sections and DEFAULT) will be modified to the new value._

NOTE: This function modifies the instance's configuration file, not
the content currently loaded into the cfg dictionary.

On the first call to modify_configfile() the content of the file is read into memory.  Successive
calls to modify_configfile() may be made, with the changes applied to the in-memory copy.  When
all changes have been applied the final call to modify_configfile() must have `save=True` to 
cause the memory version to be written out to the config file.  If the script code checks for
modifications of the config file then the modified content will be reloaded into the cfg dictionary.


### Args
`param` (str, default '')
- The param name, if modifying an existing param or adding a new param

`value` (any, default '')
- The new value to be applied to an existing param, or an added param
- Any comment text (after a '#') in the new value will be prepended to any existing comment text

`remove` (bool, default False)
- If True, the `param` config file line is removed from the config file

`add_if_not_existing` (bool, default False)
- Modify an existing param line, or add at the bottom of the config file if it is not existing
- To add a blank line leave out both `param` and `value`, or set both the `""`
- To add a comment line specify the comment in the `param` field (eg, `my_config.modify_configfile("# My comment")`)

`save` (bool, default False)
- Write the modified config file content back to the file
- `save=True` may be specified on the last modification call or an a standalone call.


### Returns
- None
- Warning messages are logged for attempting to modify or remove a non-existing param.


### Behaviors and rules
- **How modify_config works with multi-line params -** If a multi-line param is modified, the new value is
written out on a single line, and the continuation lines for the original definition remain in place.
This effectively turns the continuation lines into a new param definition, which is usually benign.  Check
for param name conflicts.
- NOTE:  In some circumstances the OS-reported timestamp for the modified config file may be erratic.
It may be necessary to add a `time.sleep(0.5)` delay between saving the modified config and the loadconfig()
reload call to avoid multiple config reloads.
        """

        line_format_re = re.compile(r'(\s*)([^\s=:]+)([\s=:]+)([^#]+)(.*)')

        if self.config_file is None:
            raise ConfigError ("Config file is None. Cannot modify config not loaded from a file.")

        if self.config_content == '':
            self.config_content = self.config_full_path.read_text()
        found_param = False
        updated_content = ''
        value = str(value)

        for line in self.config_content.split('\n'):
            out = line_format_re.match(line)
            if out:
                # print (f"1: <{out.group(1)}>")    # Any leading whitespace
                # print (f"2: <{out.group(2)}>")    # param
                # print (f"3: <{out.group(3)}>")    # whitespace, '=', ':' between param and value
                # print (f"4: <{out.group(4)}>")    # value, with trailing whitespace
                # print (f"5: <{out.group(5)}>")    # comment

                if out.group(2) != param:
                    updated_content += line + '\n'
                else:
                    found_param = True
                    if remove == True:
                        continue                            # just don't save the line

                    # Update value for the current line
                    len_current_whitespace = len(out.group(4)) - len(out.group(4).strip())
                    len_new_whitespace = len_current_whitespace - (len(value) - len(out.group(4).strip()))
                    if len_new_whitespace < 1:
                        len_new_whitespace = 1
                    if len_current_whitespace == 0:
                        len_new_whitespace = 0
                    updated_content += out.group(1) + out.group(2) + out.group(3) + value + ' '*len_new_whitespace + out.group(5) + '\n'
            else:
                updated_content += line + '\n'

        if found_param == False:
            if add_if_not_existing == True:
                updated_content += f"{param}    {value}\n"
            elif param==''  and  value==''  and  save==True: # Save-only call
                pass
            else:
                logging.warning (f"Modification of param <{param}> failed - not found in config file.  Modification skipped.")

        self.config_content = updated_content[:-1]          # Avoid adding extra \n at end of file

        if save:
            self.config_full_path.write_text(self.config_content)
            self.config_content = ''


#=====================================================================================
#=====================================================================================
#  w r i t e
#=====================================================================================
#=====================================================================================

    def write(self, savefile):
        """
## write (savefile) - Write config data to a file

***config_item() class member function***


### Arg
`savefile` (Path or str)
- Path to the output file.
- The config data will be written to an absolute path, or relative to the `core.tool.config_dir`


### Returns
- None on success
- Raises ConfigError if unable to write the file


### Behaviors and rules
- The created config file is as loaded in memory.  Any imports in the originally loaded config file
 are merged into the top-level.
        """

        cfg_list = ''
        for key in self.cfg:
            if key not in self.sections_list:
                if type(self.cfg[key]) is str  and  '#' in self.cfg[key]:
                    cfg_list += f"{key:20} = '''{self.cfg[key]}'''\n"
                else:
                    cfg_list += f"{key:20} = {self.cfg[key]}\n"
        cfg_list += '\n[DEFAULT]\n'
        for key in self.defaults:
            cfg_list += f"{key:20} = {self.defaults[key]}\n"
        for section in self.sections_list:
            cfg_list += f'\n[{section}]\n'
            for key in self.cfg[section]:
                cfg_list += f"{key:20} = {self.cfg[section][key]}\n"
        
        outfile = mungePath(savefile, core.tool.config_dir).full_path

        try:
            run_with_timeout(outfile.write_text, cfg_list, rwt_ntries=RWT_NTRIES, rwt_timeout=RWT_TIMEOUT)
        except Exception as e:
            raise ConfigError (f"Failed to write config {self.config_file} to file {outfile}\n  {type(e).__name__}: {e}")


#=====================================================================================
#=====================================================================================
#  s e c t i o n s
#=====================================================================================
#=====================================================================================

    def sections(self):
        """
## sections () - Return a list of sections in the cfg dictionary

***config_item() class member function***

For compatibility with the standard library configparser.  Also available via `<config>.sections_list`.

Example:
```
code:
    print (my_config.sections())

output:
    ['Bad params', 'SMTP']
```
        """

        return self.sections_list


#=====================================================================================
#=====================================================================================
#  c l e a r
#=====================================================================================
#=====================================================================================

    def clear(self, section=''):
        """
## clear (section='') - Purge a portion of the cfg dictionary

***config_item() class member function***

### Args
`section` (str, default '')
- `section = ''` clears the entire cfg dictionary, including all sections and DEFAULT
- `section = '<section_name>'` clears just that section
- `section = 'DEFAULT'` clears just the DEFAULT section


### Returns
- None
- A ConfigError is raised if attempting to remove a non-existing section
        """

        if section == '':
            self.cfg.clear()
            self.sections_list = []
            self.defaults.clear()
        elif section in self.sections_list:
            self.cfg.pop(section, None)
            self.sections_list.remove(section)
        elif section == 'DEFAULT':
            self.defaults.clear()
        else:
            raise ConfigError (f"Failed attempt to remove non-existing section <{section}> from config")


#=====================================================================================
#=====================================================================================
#  _ _ r e p r _ _
#=====================================================================================
#=====================================================================================

    def __repr__(self):
        stats = ""
        stats += f"\nStats for config file <{self.config_file}>:\n"
        stats += f".config_file            :  {self.config_file}\n"
        stats += f".config_dir             :  {self.config_dir}\n"
        stats += f".config_full_path       :  {self.config_full_path}\n"
        stats += f".config_timestamp       :  {self.config_timestamp}\n"
        stats += f".sections_list          :  {self.sections_list}\n"
        stats += f".force_str              :  {self.force_str}\n"
        stats += f".secondary_config       :  {self.secondary_config}\n"
        stats += f"core.tool.log_dir_base  :  {core.tool.log_dir_base}\n"
        return stats


#=====================================================================================
#=====================================================================================
#  d u m p
#=====================================================================================
#=====================================================================================

    def dump(self):
        """
## dump () - Return the formatted content of the cfg dictionary

***config_item() class member function***


### Returns
- str type pretty formatted content of the cfg dictionary, along with any sections and defaults
        """

        cfg_list = "***** Section [] *****\n"
        for key in self.cfg:
            if key not in self.sections_list:
                cfg_list += f"{key:>20} = {self.cfg[key]}  {type(self.cfg[key])}\n"
        for section in self.sections_list:
            cfg_list += f"***** Section [{section}] *****\n"
            for key in self.cfg[section]:
                cfg_list += f"{key:>20} = {self.cfg[section][key]}  {type(self.cfg[section][key])}\n"
        cfg_list += f"***** Section [DEFAULT] *****\n"
        for key in self.defaults:
            cfg_list += f"{key:>20} = {self.defaults[key]}  {type(self.defaults[key])}\n"
        return cfg_list[:-1]    # drop final '\n'





#!/usr/bin/env python3
"""cjnfuncs.SMTP - Send text message notifications and emails
"""

#==========================================================
#
#  Chris Nelson, 2018-2025
#
#==========================================================

import time
import datetime
import sys
import importlib
import re
import smtplib
from email.mime.text import MIMEText
from email.utils import formatdate
from pathlib import Path
import dkim

from .core      import logging, SndEmailError
from .mungePath import mungePath, check_path_exists
from .timevalue import timevalue
import cjnfuncs.core as core

# Configs / Constants
SND_EMAIL_NTRIES =      3           # Number of tries to send email before aborting
RETRY_WAIT =            '2s'        # seconds between retries
SERVER_TIMEOUT =        '2s'        # server connection timeout

COUNTRY_CODE =          1           # US/Canada
PHONE_NUM_LENGTH =      10          # US/Canada

# Logging events within this module are at the DEBUG level.  With this module's child logger set to
# a minimum of WARNING level by default, then logging from this module is effectively disabled.  To enable
# logging from this module add this within your tool script code:
#       logging.getLogger('cjnfuncs.smtp').setLevel(logging.DEBUG)
smtp_logger = logging.getLogger('cjnfuncs.smtp')
smtp_logger.setLevel(logging.WARNING)


#=====================================================================================
#=====================================================================================
#  s n d _ n o t i f
#=====================================================================================
#=====================================================================================

def snd_notif(subj='Notification message', msg='', urls_list=[], to='NotifList', log=False, smtp_config=None):
    """
## snd_notif (subj='Notification message', msg='', urls_list=[], to='NotifList', log=False, smtp_config=None) - Send a text message using info from the config file

Intended for use the mobile provider's (carrier's) email-to-SMS gateway email address, eg, 
`5405551212@vzwtxt.com` for Verizon, but any email address will work.

The `to` string may be the name of a config param (who's value is one or more email addresses, default 
"NotifList"), or a string with one or more email addresses. Using a config param name allows for customizing the
`to` addresses without having to edit the code.

The message to send is passed in the `msg` arg as a text string.
Three attempts are made to send the message.

    
### Args
`subj` (str, default 'Notification message')
- Text message subject field
- Some SMS/MMS apps display the subj field in bold, some in raw form, and some not at all.

`msg` (str, default '')
- Text message body

`urls_list` (list, default [])
- A list of url strings to be passed to the message sending plugin module, which should pass them to the messaging service.
- If not using a messaging service then this list is discarded, in which case include the URLs in the `msg` body.

`to` (str, default 'NotifList')
- To whom to send the message. `to` may be either an explicit string list of email addresses
(whitespace or comma separated) or a config param name (also listing one
or more whitespace or comma separated email addresses).  If the `to` arg does not
contain an '@' it is assumed to be a config param name.
- Define `NotifList` in the config file to use the default `to` value.

`log` (bool, default False)
- If True, logs that the message was sent at the WARNING level (using the root logger). If False, logs 
at the DEBUG level (using the 'cjnfuncs.smtp' logger). Useful for eliminating separate logging messages in the tool script code.
The `subj` field and `msg` body are included in the log message.

`smtp_config` (config_item class instance)
- config_item class instance containing the [SMTP] section and related params


### config dictionary params in the [SMTP] section, in addition to the config dictionary params required for snd_email

`NotifList` (optional)
- string list of email addresses (whitespace or comma separated).  
- Defining `NotifList` in the config is only required if any call to `snd_notif()` uses this
default `to` arg value.

`DontNotif` (default False)
- If True, notification messages are not sent. `log` is still honored. Useful for debug.
- Setting `DontEmail` True also blocks sending notification messages if using a carrier email-to-SMS gateway 
(not using a messaging service).  If using a messaging service then `DontEmail` has no effect.

`Msg_Handler` (str, absolute path or package.module, default None)
- If using a messaging service, such as Twilio, this param declares the path to the message sending plugin module.  
- The module must implement a `sender()` function, which will be called with a `package` dictionary containing `subj`, `msg`, `urls`, and `to` key:value pairs, and
  a reference to the `smtp_config`.  See the [SMTP.md](https://github.com/cjnaz/cjnfuncs/blob/main/SMTP.md) for an example Msg_Handler module implementation.
- `Msg_Handler` may be a full absolute path to a Python module (eg, `/path-to-module/twilioSender.py`), or an installed package.module reference (eg, `mypackage.twilioSender`).

`country_code` (int or str, default 1 (US/Canada), required only if `Msg_Handler` is defined in the config [SMTP] section)
- Number without a preceding '+', eg '1' for US/Canada phone numbers
- If `get_type='numbers'` then each phone number is prepended with `+` plus `country_code` (eg, '+1'), but only if the number does not already have a country code.
- If a `to` phone number has a different country code, it is retained.

`number_length` (int, default 10 (US/Canada), required only if `Msg_Handler` is defined in the config [SMTP] section)
- The number of digits in a valid phone number (not including the country code), eg 10 for US/Canada phone numbers


### Returns
- None
- Raises `SndEmailError` on error


### Behaviors and rules
- `snd_notif()` uses `snd_email()` to send the message (if not using a messaging service). See `snd_email()` for related setup.
    """

    if smtp_config.getcfg('DontNotif', fallback=False, section='SMTP'):
        if log:
            logging.warning (f"Notification NOT sent <{subj}> <{msg}>")
        else:
            smtp_logger.debug (f"Notification NOT sent <{subj}> <{msg}>")
        return


    msg_handler = smtp_config.getcfg('Msg_Handler', fallback=None, section='SMTP')
    if msg_handler:
        # Import the messaging service handler
        if msg_handler.startswith('/'):                 # Absolute path case
            if not check_path_exists(msg_handler):
                raise FileNotFoundError (f"Can't find SMS/MMS message handler <{msg_handler}>")
            if msg_handler.endswith('.py'):
                msg_handler = msg_handler[:-3]
            xx = mungePath(msg_handler)
            xx_parent = str(xx.parent)
            if xx_parent not in sys.path:
                sys.path.append(xx_parent)
            try:
                # smtp_logger.info (xx)
                sender_plugin = __import__(xx.name)
            except Exception as e:
                raise ImportError (f"Can't import SMS/MMS message handler <{msg_handler}>\n  {e}")
            smtp_logger.debug (f"Imported message sender plugin <{msg_handler}>, version <{sender_plugin.__version__}>")
        else:                                           # package.module case
            try:
                sender_plugin = importlib.import_module(msg_handler)
            except Exception as e:
                raise ImportError (f"Can't import SMS/MMS message handler <{msg_handler}>\n  {e}")
            smtp_logger.debug (f"Imported message sender plugin <{sender_plugin.__name__}>, version <{sender_plugin.__version__}>")

        package = {'subj': subj,
                   'msg':  msg,
                   'urls': urls_list,
                   'to':   list_to(to, 'numbers', subj=subj, smtp_config=smtp_config) }

        sender_plugin.sender(package, smtp_config)

    else:
        snd_email (subj=subj, body=msg, to=to, smtp_config=smtp_config)


    if log:
        logging.warning (f"Notification sent <{subj}> <{msg}>")
    else:
        smtp_logger.debug (f"Notification sent <{subj}> <{msg}>")


#=====================================================================================
#=====================================================================================
#  s n d _ e m a i l
#=====================================================================================
#=====================================================================================

def snd_email(subj, to, body=None, filename=None, htmlfile=None, log=False, smtp_config=None):
    """
## snd_email (subj, to, body=None, filename=None, htmlfile=None, log=False, smtp_config=None) - Send an email message using info from the config file

The `to` string may be the name of a config param (who's value is one or more email addresses),
or a string with one or more email addresses. Using a config param name allows for customizing the
`to` addresses without having to edit the code.

What to send may be a `body` string, the text contents of `filename`, or the HTML-formatted contents
of `htmlfile`, in this order of precedent.  MIME multi-part is not supported.

DKIM signing is optionally supported.

Three attempts are made to send the message (see `EmailNTries`, below).


### Args
`subj` (str)
- Email subject text

`to` (str)
- To whom to send the message. `to` may be either an explicit string list of email addresses
(whitespace or comma separated) or a config param name in the [SMTP] section (also listing one
or more whitespace or comma separated email addresses).  If the `to` arg does not
contain an '@' it is assumed to be a config param name.

`body` (str, default None)
- A string message to be sent

`filename` (str, default None)
- A str or Path to the file who's content will be sent as the body of the message
- The file path is relative to the `core.tool.cache_dir`, or an absolute path

`htmlfile` (str, default None)
- A str or Path to the file who's HTML-formatted content will be sent as the body of the message
- The file path is relative to the `core.tool.cache_dir`, or an absolute path

`log` (bool, default False)
- If True, logs that the message was sent at the WARNING level (using the root logger). If False, logs 
at the DEBUG level (using the 'cjnfuncs.smpt' logger). Useful for eliminating separate logging messages in the tool script code.
The `subj` field is part of the log message.

`smtp_config` (config_item class instance)
- config_item class instance containing the [SMTP] section and related params


### config dictionary params in the [SMTP] section
`EmailFrom`
- An email address, such as `me@myserver.com`

`EmailServer`
- The SMTP server name, such as `mail.myserver.com`

`EmailServerPort`
- The SMTP server port (one of `P25`, `P465`, `P587`, or `P587TLS`)

`EmailUser`
- Username for `EmailServer` login, if required by the server

`EmailPass`
- Password for `EmailServer` login, if required by the server

`DontEmail` (default False)
- If True, messages are not sent. Useful for debug. Also blocks `snd_notif()` messages not sent thru a messaging service.

`EmailVerbose` (default False)
- If True, detailed transactions with the SMTP server are sent to stdout. Useful for debug.

`EmailNTries` (type int, default 3)
- Number of tries to send email before aborting

`EmailRetryWait` (seconds, type int, float, or timevalue, default 2s)
- Number of seconds to wait between retry attempts

`EmailServerTimeout` (seconds, type int, float, or timevalue, default 2s)
- Server connection timeout

`EmailDKIMDomain` (required if using DKIM email signing)
- The domain of the public-facing SMTP server, eg `mydomain.com`
- Defining `EmailDKIMDomain` enables DKIM signing, and also requires `EmailDKIMPem` and `EmailDKIMSelector`

`EmailDKIMPem` (required if using DKIM email signing)
- Full path to the private key file of the public-facing SMTP server at the `EmailDomain`, eg `/home/me/creds_mydomain.com.pem`
- Make sure this file is readable only to the user
- You may be able to obtain this key in cPanel for your shared-hosting service

`EmailDKIMSelector` (required if using DKIM email signing)
- The DKIM selector string, eg 'default'


### Returns
- None
- Raises SndEmailError on error


### Behaviors and rules
- One of `body`, `filename`, or `htmlfile` must be specified. Looked for in this order, and the first 
found is used.
- EmailServerPort must be one of the following:
  - P25:  SMTP to port 25 without any encryption
  - P465: SMTP_SSL to port 465
  - P587: SMTP to port 587 without any encryption
  - P587TLS:  SMTP to port 587 and with TLS encryption
- It is recommended (not required) that the email server params be placed in a user-read-only
file in the user's home directory, such as `~/creds_SMTP`, and imported by the main config file.
Some email servers require that the `EmailFrom` address be of the same domain as the server, 
so it may be practical to bundle `EmailFrom` with the server specifics.  Place all of these in 
`~/creds_SMTP`:
  - `EmailFrom`, `EmailServer`, `EmailServerPort`, `EmailUser`, and `EmailPass`
  - If DKIM signing is used, also include `EmailDKIMDomain`, `EmailDKIMPem`, and `EmailDKIMSelector`
- `snd_email()` does not support multi-part MIME (an html send wont have a plain text part).
- Checking the validity of email addresses is very basic... an email address must contain an '@'.
    """

    if smtp_config is None:
        raise SndEmailError ("smtp_section required for SMTP params")


    # Deal with what to send
    if body:
        msg_type = "plain"
        m_text = body

    elif filename:
        xx = mungePath(filename, core.tool.cache_dir)
        try:
            msg_type = "plain"
            with Path.open(xx.full_path) as ifile:
                m_text = ifile.read()
        except Exception as e:
            raise SndEmailError (f"snd_email - Message subject <{subj}>:  Failed to load <{xx.full_path}>.\n  {e}") from None

    elif htmlfile:
        xx = mungePath(htmlfile, core.tool.cache_dir)
        try:
            msg_type = "html"
            with Path.open(xx.full_path) as ifile:
                m_text = ifile.read()
        except Exception as e:
            raise SndEmailError (f"snd_email - Message subject <{subj}>:  Failed to load <{xx.full_path}>.\n  {e}") from None

    else:
        raise SndEmailError (f"snd_email - Message subject <{subj}>:  No body, filename, or htmlfile specified.")

    m_text += ('\n' + datetime.datetime.now().astimezone().strftime("%a %b %d %Y - %H:%M:%S"))


    # Deal with 'to'
    To = list_to(to, 'emails', subj, smtp_config=smtp_config)


    # Gather, check remaining config params
    ntries =            smtp_config.getcfg('EmailNTries', SND_EMAIL_NTRIES, types=int, section='SMTP')
    retry_wait =        timevalue(smtp_config.getcfg('EmailRetryWait', RETRY_WAIT, types=[int, float, str], section='SMTP')).seconds
    server_timeout =    timevalue(smtp_config.getcfg('EmailServerTimeout', SERVER_TIMEOUT, types=[int, float, str], section='SMTP')).seconds
    email_from =        smtp_config.getcfg('EmailFrom', types=str, section='SMTP')
    cfg_server =        smtp_config.getcfg('EmailServer', types=str, section='SMTP')
    cfg_port =          smtp_config.getcfg('EmailServerPort', types=str, section='SMTP').lower()
    if cfg_port not in ['p25', 'p465', 'p587', 'p587tls']:
        raise SndEmailError (f"snd_email - Config EmailServerPort <{cfg_port}> is invalid")

    email_user =        str(smtp_config.getcfg('EmailUser', None, types=[str, int, float], section='SMTP')) # username may be numeric - optional
    if email_user:
        email_pass =    str(smtp_config.getcfg('EmailPass', types=[str, int, float], section='SMTP'))      # password may be numeric - required if EmailUser provided

    dkim_domain =       smtp_config.getcfg('EmailDKIMDomain', None, types=str, section='SMTP')
    if dkim_domain:
        dkim_pem =      smtp_config.getcfg('EmailDKIMPem', None, types=str, section='SMTP')
        if not dkim_pem:
            raise SndEmailError (f"snd_email - Config <EmailDKIMPem> is required for SMTP DKIM signing")
        dkim_selector = smtp_config.getcfg('EmailDKIMSelector', None, types=str, section='SMTP')
        if not dkim_selector:
            raise SndEmailError (f"snd_email - Config <EmailDKIMSelector> is required for SMTP DKIM signing")


    # Send the message, with retries
    for trynum in range(ntries):
        try:
            msg = MIMEText(m_text, msg_type)
            msg['Subject'] = subj
            msg['From']    = email_from
            msg['To']      = ", ".join(To)
            msg["Date"]    = formatdate(localtime=True)

            # smtp_logger.debug (msg)

            if smtp_config.getcfg('DontEmail', fallback=False, types=bool, section='SMTP'):
                if log:
                    logging.warning (f"Email NOT sent <{subj}>")
                else:
                    smtp_logger.debug (f"Email NOT sent <{subj}>")
                return

            # Add DKIM signature if EmailDKIMDomain is specified
            if dkim_domain:
                privateKey = Path(dkim_pem).read_text()
                sig = dkim.sign(message=msg.as_bytes(),
                                selector=   bytes(dkim_selector, 'UTF8'),
                                domain=     bytes(dkim_domain, 'UTF8'),
                                privkey=    bytes(privateKey, 'UTF8'),
                                include_headers= ['from', 'to', 'subject', 'date'])
                sig = sig.decode()
                msg['DKIM-Signature'] = sig[len("DKIM-Signature: "):]

            smtp_logger.debug (f"Initialize the SMTP server connection for port <{cfg_port}>")
            if cfg_port == "p25":
                server = smtplib.SMTP(cfg_server, 25, timeout=server_timeout)
            elif cfg_port == "p465":
                server = smtplib.SMTP_SSL(cfg_server, 465, timeout=server_timeout)
            else: # cfg_port == "p587" or "p587tls"
                server = smtplib.SMTP(cfg_server, 587, timeout=server_timeout)

            if smtp_config.getcfg("EmailVerbose", False, types=[bool], section='SMTP'):
                smtp_logger.debug ("Set SMTP connection debuglevel(1)")
                server.set_debuglevel(1)

            if cfg_port == "p587tls":
                smtp_logger.debug ("Start TLS")
                server.starttls()

            # if email_user:
            if cfg_port.startswith('p587'):
                smtp_logger.debug (f"Logging into SMTP server")
                server.login (email_user, email_pass)

            smtp_logger.debug (f"Sending message <{subj}>")
            server.sendmail(email_from, To, msg.as_string())
            server.quit()

            if log:
                logging.warning (f"Email sent <{subj}>")
            else:
                smtp_logger.debug (f"Email sent <{subj}>")
            return

        except Exception as e:
            last_error = e
            if trynum < ntries -1:
                smtp_logger.debug(f"Email send try {trynum} failed.  Retry in <{retry_wait} sec>:\n  <{e}>")
                time.sleep(retry_wait)
            continue

    raise SndEmailError (f"snd_email:  Send failed for <{subj}>:\n  <{last_error}>")


#=====================================================================================
#=====================================================================================
#  l i s t _ t o
#=====================================================================================
#=====================================================================================

def list_to(raw_in, get_type, subj, smtp_config):
    """
### list_to(raw_in, get_type, subj, smtp_config) - Build list of phone numbers or email addresses from raw_in

`list_to` handles several translations, controlled by `get_type` ('numbers' or 'emails'), including dereferencing through
config param, extracting and constructing a proper phone number list for messaging services (eg, Twilio), and constructing an 
email address list.

### Args
`raw_in` (str or int (in the case of a single phone number))
- A string list of phone numbers or email addresses, separated by either whitespace or ','
- If the value is the name of a param in the smtp_config [SMTP] section then that param's value is used, else `raw_in` is directly parsed
- Phone numbers may optionally include '+<country_code>', eg +14325551212

`get_type` (str)
- 'numbers' - a list of 1 or more phone numbers will be returned
  - '+<country_code>', eg '+1' is prepended to each number, if not provided
  - If the `raw_in` is one or more email-to-SMS gateway addresses (eg, 4805551212@vzwpix.com), then the local-part (the part before '@') is extracted as a phone number.
  This feature enables a config parameter containing a list of carrier email-to-SMS gateway addresses to be directly used with a SMS/MMS messaging service such as Twilio.
- 'emails' - a list of 1 or more email addresses will be returned
  - Address validity checking is minimal:  Email addresses must contain an '@'.

`subj` (str)
- The notification or email subject field - used only for raised errors to aid debug tracing

`smtp_config` (config_item class instance)
- config_item class instance containing the [SMTP] section and related params


### config dictionary params in the [SMTP] section
`country_code` (int or str, default 1 (US/Canada), required only if `get_type='numbers'`)
- Number without a preceding '+', eg '1' for US/Canada phone numbers
- If `get_type='numbers'` then each phone number is prepended with `+` plus `country_code` (eg, '+1'), but only if the number does not already have a country code.
- If a `raw_in` phone number has a different country code, it is retained.

`number_length` (int, default 10 (US/Canada), required only if `get_type= 'numbers'`)
- The number of digits in a valid phone number (not including the country code) for the given `country_code`, eg 10 for US/Canada phone numbers

### Returns
- Either a list of phone numbers or a list of email addresses, eg:
  - ['+14325551212']
  - ['+14325551212', '+14802345678']
  - ['4325551212@txt.att.net', '4802345678@vxwpix.com', 'myemail@tmomail.net']
- Raises SndEmailError on any errors
    """

    xx = smtp_config.getcfg(raw_in, fallback=None, section='SMTP')
    if xx:              # raw_in is a config-defined param (of 1 or more phone numbers)
        raw_in = xx

    items =     re.split(r'[,\s]+', str(raw_in))
    cc =        None    # Force load once per call
    to_list =   []

    for item in items:
        if get_type == 'numbers':
            if not cc:
                cc =        '+' + str(smtp_config.getcfg('country_code', COUNTRY_CODE, section='SMTP'))     # '+1'
                num_len =   smtp_config.getcfg('number_length', PHONE_NUM_LENGTH, section='SMTP')           # 10
            if '@' in item:
                num_part = item.split('@')[0]
            else:
                num_part = item

            if num_part.startswith(cc):                                 # Strip optional default country code prefix
                num_part = num_part.replace(cc,'')

            if num_part.startswith('+')  and  num_part[1:].isdigit():   # Non-default country code
                to_list.append(num_part)
            elif num_part.isdigit()  and  len(num_part) == num_len:     # No country code prefix remaining
                to_list.append(cc + num_part)
            else:
                raise SndEmailError (f"Message subject <{subj}>:  <{num_part}> is not a valid phone number")

        elif get_type == 'emails':
            if '@' not in item:
                raise SndEmailError (f"Message subject <{subj}>:  <{item}> is not a valid email address")
            to_list.append(item)
        
        else:
            raise SndEmailError (f"Message subject <{subj}>:  <{get_type}> is not a valid get_type for list_to()")

    return (to_list)



#!/usr/bin/env python3
"""cjnfuncs - Establish the base environment for cjnfuncs, and logging related support
"""

#==========================================================
#
#  Chris Nelson, 2018-2025
#
# TODO Identification of the main module / tool script file may be wrong if a script imports a script that then imports cjnfuncs.
#==========================================================

import sys
import logging
import inspect
import platform
from pathlib import Path
import __main__
import appdirs
import datetime


# Configs / Constants
# FILE_LOGGING_FORMAT    = '{asctime}/{module}/{funcName}/{levelname}:  {message}'    # Classic format
FILE_LOGGING_FORMAT    = '{asctime} {module:>15}.{funcName:20} {levelname:>8}:  {message}'
CONSOLE_LOGGING_FORMAT = '{module:>15}.{funcName:20} - {levelname:>8}:  {message}'


# Get the main / calling module info (who imported cjnfuncs.core)
stack = inspect.stack()
calling_module = ""
for item in stack:  # Look for the import cjnfuncs line
    code_context = item[4]
    if code_context is not None:
        if "cjnfuncs" in code_context[0]:
            calling_module = inspect.getmodule(item[0])
            break


cjnfuncs_logger = logging.getLogger('cjnfuncs').setLevel(logging.WARNING)


#=====================================================================================
#=====================================================================================
#  M o d u l e   e x c e p t i o n s
#=====================================================================================
#=====================================================================================

class Error(Exception):
    """Base class for exceptions in this module."""
    pass

class ConfigError(Error):
    """Exceptions raised for config file function errors.
    Attributes:
        message -- error message including item in error
    Format:
        ConfigError:  <function> - <message>.
    """
    def __init__(self, message):
        self.message = message

class SndEmailError(Error):
    """Exceptions raised for snd_email and snd_notif errors.
    Attributes:
        message -- error message including item in error
    Format:
        SndEmailError:  <function> - <message>.
    """
    def __init__(self, message):
        self.message = message


#=====================================================================================
#=====================================================================================
#  C l a s s   s e t _ t o o l n a m e
#=====================================================================================
#=====================================================================================

class set_toolname():
    """
## Class set_toolname (toolname) - Set target directories for config and data storage

set_toolname() centralizes and establishes a set of base directory path variables for use in
the tool script.  It looks first for existing directories, based on the specified toolname, in
the site-wide (system-wide) locations and then in user-specific locations.  Specifically, site-wide 
config and/or data directories are looked for at `/etc/xdg/<toolname>` and/or 
`/usr/share/<toolname>`.  If site-wide directories are not 
found then the user-specific environment is assumed.  No directories are created.


### Args
`toolname` (str)
- Name of the tool


### Returns
- Handle to the `set_toolname()` instance.
- The handle is also stored in the `core.tool` global.  See examples for proper access.


### Behaviors, rules, and _variances from the XDG spec and/or the appdirs package_
- For a `user` setup, the `.log_dir_base` is initially set to the `.user_data_dir` (a variance from XDG spec).
If a config file is subsequently
loaded then the `.log_dir_base` is changed to the `.user_config_dir`.  (Not changed for a `site` setup.)
Thus, for a `user` setup, logging defaults to the configuration directory.  This is a 
style variance, and can be disabled by setting `remap_logdirbase = False` on the config_item instantiation, or
by reassigning: `core.tool.log_dir_base = core.tool.user_log_dir` (or any
other directory) before a subsequent call to `loadconfig()` or `setuplogging()`.
(The XDG spec says logging goes to the `.user_state_dir`, while appdirs sets it to the `.user_cache_dir/log`.)

- The last operation within `set_toolname()` is to call `setuplogging()`, which initializes the root
logger to log to the console.  The `.log_dir` and `.log_file` attributes are set to `None`, while 
`.log_full_path` is set to `__console__`.  Having `set_toolname()` call `setuplogging()` ensures that any logging events
before a user-code call to `setuplogging()` or `loadconfig()` are properly logged.

- For a `site` setup, the `.site_data_dir` is set to `/usr/share/<toolname>`.  The XDG spec states that 
the `.cache_dir` and `.state_dir` should be in the root user tree; however, set_toolname() sets these two 
also to the `.site_data_dir`.

    """
    def __init__(self, toolname):
        global tool             # Tool info is accessed via core.tool...
        tool = self
        self.toolname  = toolname
        self.main_module        = calling_module
        try:
            self.main_full_path = Path(self.main_module.__file__)
            self.main_dir       = self.main_full_path.parent
        except:                 # Referencing an installed module's file/path may not be legal
            self.main_full_path = None
            self.main_dir       = None
        self.user_config_dir    = Path(appdirs.user_config_dir(toolname, appauthor=False))  # appauthor=False to avoid double toolname on Windows
        self.user_data_dir      = Path(appdirs.user_data_dir  (toolname, appauthor=False))
        self.user_state_dir     = Path(appdirs.user_state_dir (toolname, appauthor=False))
        self.user_cache_dir     = Path(appdirs.user_cache_dir (toolname, appauthor=False))
        self.user_log_dir       = Path(appdirs.user_log_dir   (toolname, appauthor=False))
        self.site_config_dir    = Path(appdirs.site_config_dir(toolname, appauthor=False))
        if platform.system() == "Windows":
            self.site_data_dir  = Path(appdirs.site_data_dir  (toolname, appauthor=False))
        else:   # Linux, ...
            self.site_data_dir  = Path("/usr/share") / toolname

        if self.site_config_dir.exists()  or  self.site_data_dir.exists():      # TODO hang risk, but assumed to be on a local drive
            self.env_defined= "site"
            self.config_dir     = self.site_config_dir
            self.data_dir       = self.site_data_dir
            self.state_dir      = self.site_data_dir
            self.cache_dir      = self.site_data_dir
            self.log_dir_base   = self.site_data_dir
        else:
            self.env_defined= "user"
            self.config_dir     = self.user_config_dir
            self.data_dir       = self.user_data_dir
            self.state_dir      = self.user_state_dir
            self.cache_dir      = self.user_cache_dir
            self.log_dir_base   = self.user_data_dir

        setuplogging()

    def __repr__(self):
        stats = ""
        stats +=  f"\nStats for set_toolname <{self.toolname}>:\n"
        stats +=  f".toolname         :  {self.toolname}\n"
        stats +=  f".main_module      :  {self.main_module}\n"
        stats +=  f".main_full_path   :  {self.main_full_path}\n"
        stats +=  f".main_dir         :  {self.main_dir}\n"

        stats +=  f"General user and site paths:\n"
        stats +=  f".user_config_dir  :  {self.user_config_dir}\n"
        stats +=  f".user_data_dir    :  {self.user_data_dir}\n"
        stats +=  f".user_state_dir   :  {self.user_state_dir}\n"
        stats +=  f".user_cache_dir   :  {self.user_cache_dir}\n"
        stats +=  f".user_log_dir     :  {self.user_log_dir}\n"
        stats +=  f".site_config_dir  :  {self.site_config_dir}\n"
        stats +=  f".site_data_dir    :  {self.site_data_dir}\n"

        stats +=  f"Based on found user or site dirs:\n"
        stats +=  f".env_defined      :  {self.env_defined}\n"
        stats +=  f".config_dir       :  {self.config_dir}\n"
        stats +=  f".data_dir         :  {self.data_dir}\n"
        stats +=  f".state_dir        :  {self.state_dir}\n"
        stats +=  f".cache_dir        :  {self.cache_dir}\n"
        stats +=  f".log_dir_base     :  {self.log_dir_base}\n"
        stats +=  f".log_dir          :  {self.log_dir}\n"
        stats +=  f".log_file         :  {self.log_file}\n"
        stats +=  f".log_full_path    :  {self.log_full_path}"
        return stats


#=====================================================================================
#=====================================================================================
#  s e t u p l o g g i n g
#=====================================================================================
#=====================================================================================

logger_q = None
logger_listener = None


def setuplogging(call_logfile=None, call_logfile_wins=False, config_logfile=None, ConsoleLogFormat=None, FileLogFormat=None):
    """
## setuplogging (call_logfile=None, call_logfile_wins=False, config_logfile=None, ConsoleLogFormat=None, FileLogFormat=None) - Set up the root logger

Logging may be directed to the console (stdout), or to a file.  Each time setuplogging()
is called the root logger's output path (log file or console) may be reassigned.

Calling `setuplogging()` with no args results in:
- Logging output to the console
- Logging format set to the default console logging format
- Logging level is unchanged (the Python default is 30/WARNING)

setuplogging() works standalone or in conjunction with `cjnfuncs.configman.loadconfig()`.
If a loaded config file has a `LogFile` parameter then loadconfig() passes it's value thru
`config_logfile`.  loadconfig() also passes along any `call_logfile` and `call_logfile_wins` switch
that were passed to loadconfig() from the tool script.  This mechanism allows the tool script
to override any config `LogFile`, such as for directing output to the console for a tool script's 
interactive use.  For example, this call will set logging output to the console regardless of 
the LogFile declared in the config file:

    setuplogging (call_logfile=None, call_logfile_wins=True, config_logfile='some_logfile.txt')

    
### Args
`call_logfile` (Path or str, default None)
- Potential log file passed typically from the tool script.  Selected by `call_logfile_wins = True`.
call_logfile may be an absolute path or relative to the `core.tool.log_dir_base` directory.  
- `None` specifies the console.

`call_logfile_wins` (bool, default False)
- If True, the `call_logfile` is selected.  If False, the `config_logfile` is selected.

`config_logfile` (str, default None)
- Potential log file passed typically from loadconfig() if there is a `LogFile` param in the 
loaded config.  Selected by `call_logfile_wins = False`.
config_logfile may be an absolute path or relative to the `core.tool.log_dir_base` directory.  
- `None` specifies the console.

`ConsoleLogFormat` (str, default None)
- Overrides the default console logging format: `{module:>15}.{funcName:20} - {levelname:>8}:  {message}`.
- loadconfig() passes `ConsoleLogFormat` from the primary config file, if defined.

`FileLogFormat` (str, default None)
- Overrides the default file logging format: `{asctime} {module:>15}.{funcName:20} {levelname:>8}:  {message}`.
- loadconfig() passes `FileLogFormat` from the primary config file, if defined.


### Returns
- None


### Behaviors and rules
- All cjnfuncs modules start with WARNING level logging.  To enable debug or info logging on a specific cjnfuncs 
module set the named child logger from your tool script, eg:

    logging.getLogger('cjnfuncs.rwt').setLevel(logging.DEBUG)

These modules support setting child logger levels:  configman, deployfiles, resourcelock, rwt, SMTP
    """

    from .mungePath import mungePath

    _lfp = "__console__"
    if call_logfile_wins == False  and  config_logfile:
        _lfp = mungePath(config_logfile, tool.log_dir_base)

    if call_logfile_wins == True   and  call_logfile:
        _lfp = mungePath(call_logfile, tool.log_dir_base)

    logger = logging.getLogger()

    if _lfp == "__console__":
        _fmt = CONSOLE_LOGGING_FORMAT  if ConsoleLogFormat == None  else  ConsoleLogFormat
        log_format = logging.Formatter(_fmt, style='{')
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(log_format)
        logger.handlers.clear()
        logger.addHandler(handler)

        tool.log_dir = None
        tool.log_file = None
        tool.log_full_path = "__console__"

    else:
        mungePath(_lfp.parent, mkdir=True)      # Force make the target dir     TODO Hang risk if not on local drive.  rwt?
        _fmt = FILE_LOGGING_FORMAT  if FileLogFormat == None  else  FileLogFormat
        log_format = logging.Formatter(_fmt, style='{')
        handler = logging.FileHandler(_lfp.full_path, "a")
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(log_format)
        logger.handlers.clear()
        logger.addHandler(handler)
    
        tool.log_dir = _lfp.parent
        tool.log_file = _lfp.name
        tool.log_full_path = _lfp.full_path


#=====================================================================================
#=====================================================================================
#   s e t  /  r e s t o r e _ l o g g i n g _ l e v e l  f u n c t i o n s
#=====================================================================================
#=====================================================================================

known_loggers = {}


def set_logging_level(new_level, logger_name='', clear=False, save=False):
    """
## set_logging_level (new_level, logger_name='', clear=False, save=False) - Save the current logging level and set the new_level

The current logging level is optionally saved on a stack and can be restored by a call to `restore_logging_level()`.
Calling set_logging_level is exactly equivalent to `logging.getLogger(logger).setLevel(new_level)`, with the 
added history mechanism (and a cleaner syntax).


### Args
`new_level` (int)
- The new logging level to be set.  Values may be set to the logging module defined levels, or their integer 
equivalents (or to any integer value that makes sense):  logging.DEBUG (10), logging.INFO (20), logging.WARNING (30), 
logging.ERROR (40), or logging.CRITICAL (50).

`logger` (str, default '' (root logger))
- Optional child logger name to be set, eg: 'cjnfuncs.rwt'
- If omitted, sets the root logger level

`clear` (bool, default False)
- If True, the logging level history stack is cleared

`save` (bool, default False)
- If True, the current logging level is saved to the stack for being restored by `restore_logging_level()`.
- If also clear=True, then the clear is done first, resulting in the stack having only the prior logging level.


### Returns
- None


### Behaviors
- NOTE that a child logger that has not been set to a logging level will have a logging level = 0, which Python seems to
treat the same as logging.WARNING (30).

    """

    global known_loggers
    if logger_name not in known_loggers:
        known_loggers[logger_name] = []
    if clear:
        known_loggers[logger_name] = []
    if save:
        known_loggers[logger_name].append(logging.getLogger(logger_name).level)
    logging.getLogger(logger_name).setLevel(new_level)



def restore_logging_level(logger_name=''):
    """
## restore_logging_level (logger_name='') - Restore the prior logging level from the stack

The prior saved logging level for the specified child/root logger (from the prior set_logging_level call) is popped
from the stack and set as the current logging level.
If the stack is empty then the logging level is set to logging.WARNING (30).


### Args
`logger` (str, default '' (root logger))
- Optional child logger name to be set, eg: 'cjnfuncs.rwt'
- If omitted, restores the root logger level


### Returns
- None
    """
    global known_loggers
    try:
        logging.getLogger(logger_name).setLevel(known_loggers[logger_name].pop())
    except:
        logging.getLogger(logger_name).setLevel(logging.WARNING)


def get_logging_level_stack(logger_name=''):
    """
## get_logging_level_stack (logger_name='') - Return the content of the stack

Useful for debug and testing.  The stack may be cleared with a call to `set_logging_level(clear=True)` or `pop_logging_level_stack(clear=True)`.


### Args
`logger` (str, default '' (root logger))
- Optional child logger name to be set, eg: 'cjnfuncs.rwt'
- If omitted, returns the root logger level stack


### Returns
- A list of the prior saved logging levels for the specified child/root logger, or an empty list if no
values are on the stack or have previously been saved.
"""
    global known_loggers
    try:
        return known_loggers[logger_name]
    except:
        return []
    return ll_history


def pop_logging_level_stack(logger_name='', clear=False):
    """
## pop_logging_level_stack (logger_name='', clear=False) - Discard top of the stack

Useful if the preexisting logging level was saved to the stack, but should be discarded.


### Args
`logger` (str, default '' (root logger))
- Optional child logger name to be stack popped, eg: 'cjnfuncs.rwt'
- If omitted, pops the stack of the root logger

`clear` (bool, default False)
- If True, the logging level history stack is cleared


### Returns
- Logging level stack after pop or clear
"""
    global known_loggers
    if clear:
        xx = known_loggers[logger_name] = []
    else:
        try:
            xx = known_loggers[logger_name].pop()
        except:
            xx = []               # Stack was empty or no prior history

    return xx


#=====================================================================================
#=====================================================================================
#   p e r i o d i c _ l o g
#=====================================================================================
#=====================================================================================

cats = {}

class _periodic_log:
    def __init__(self, log_interval, log_level, logger_name):

        from .timevalue import timevalue

        self.next_dt = datetime.datetime.now()
        self.log_interval = datetime.timedelta(seconds=timevalue(log_interval).seconds)
        self.log_level = log_level
        self.logger_name = logger_name

    def plog(self, message, log_level, cat):
        now_dt = datetime.datetime.now()
        if now_dt > self.next_dt:
            if log_level is None:
                log_level = self.log_level
            _logger = logging.getLogger(self.logger_name)
            _logger.log(log_level, f"[PLog-{cat}] {message}")
            self.next_dt = now_dt + self.log_interval


def periodic_log(message, category='Cat1', logger_name='', log_interval='10m', log_level=None):
    """
## periodic_log (message, category='Cat1', logger_name='', log_interval='10m', log_level=30) - Log a message infrequently

Log infrequently so as to avoid flooding the log.  The `category` arg provides for independent
log intervals for different types of logging events.


### Args

`message` (str)
- The message text to be logged
- Only logged if the first time for this `category` or the log_interval has expired

`category` (int or str, default 'Cat1')
- Allows for multiple, independent concurrent periodic_log streams
- `category` should typically be an int or str.  Used as a dict key.

`logger_name` (str, default '')
- Name of child or root logger.  Only remembered on the first log call
for this category (ignored of subsequent calls).

`log_interval` (timevalue, default '10m')
- How often this category's messages will be logged.  Only remembered on the first log call
for this category (ignored of subsequent calls).

`log_level` (int, default logging.WARNING/30)
- The default for this category is set on first call
- This default value may be overridden on subsequent calls for this category


### Returns
- None
      """
    if category not in cats:
        if log_level is None:
            log_level = logging.WARNING
        xx = _periodic_log(log_interval, log_level, logger_name)
        cats[category] = xx
    
    p_logger = cats[category]
    p_logger.plog(message, log_level, category)



#!/usr/bin/env python3
"""cjnfuncs.timevalue - Shorthand time values and related time management functions
"""

import datetime

#==========================================================
#
#  Chris Nelson, 2018-2025
#
#==========================================================


#=====================================================================================
#=====================================================================================
#  C l a s s   t i m e v a l u e
#=====================================================================================
#=====================================================================================

class timevalue():
    def __init__(self, orig_val):
        """
## Class timevalue (orig_val) - Convert time value strings of various resolutions to seconds

`timevalue()` provides a convenience mechanism for working with time values and time/datetime calculations.
timevalues are generally an integer or float value with an attached single character time resolution, such as "5m".
`timevalue()` also accepts integer and float values, which are interpreted as seconds resolution. Also see retime().


### Args
`orig_val` (str, int, or float)
- The original, passed-in value
- str values support these case insensitive suffix characters: 's'econds, 'm'inutes, 'h'ours, 'd'ays, and 'w'eeks
- int and float values (and str values without suffix characters) have assumed resolution seconds (eg, 7.5 means 7.5 seconds)

### Returns
- Handle to instance
- Raises ValueError if the numeric value cannot be parsed or if given an unsupported time unit suffix.


### Instance attributes
- `.orig_val` - orig_val value passed in, type str (converted to str if int or float passed in)
- `.seconds` - time value in seconds resolution, type float, useful for time calculations
- `.unit_char` - the single character suffix unit of the `orig_val` value.  's' for int and float orig_val values.
- `.unit_str` - the long-form units of the `orig_val` value useful for printing/logging ("secs", "mins", "hours", "days", or "weeks")
        """
        self.orig_val = str(orig_val)

        if type(orig_val) in [int, float]:              # Case int or float
            self.seconds =  float(orig_val)
            self.unit_char = "s"
            self.unit_str =  "secs"
        else:
            try:
                self.seconds = float(orig_val)          # Case str without units
                self.unit_char = "s"
                self.unit_str = "secs"
                return
            except:
                pass
            self.unit_char =  orig_val[-1:].lower()     # Case str with units
            if self.unit_char == "s":
                self.seconds =  float(orig_val[:-1])
                self.unit_str = "secs"
            elif self.unit_char == "m":
                self.seconds =  float(orig_val[:-1]) * 60
                self.unit_str = "mins"
            elif self.unit_char == "h":
                self.seconds =  float(orig_val[:-1]) * 60*60
                self.unit_str = "hours"
            elif self.unit_char == "d":
                self.seconds =  float(orig_val[:-1]) * 60*60*24
                self.unit_str = "days"
            elif self.unit_char == "w":
                self.seconds =  float(orig_val[:-1]) * 60*60*24*7
                self.unit_str = "weeks"
            else:
                raise ValueError(f"Illegal time units <{self.unit_char}> in time string <{orig_val}>")

    def stats(self):
        return self.__repr__()

    def __repr__(self):
        stats = ""
        stats +=  f".orig_val   :  {self.orig_val:8} {type(self.orig_val)}\n"
        stats +=  f".seconds    :  {self.seconds:<8} {type(self.seconds)}\n"
        stats +=  f".unit_char  :  {self.unit_char:8} {type(self.unit_char)}\n"
        stats +=  f".unit_str   :  {self.unit_str:8} {type(self.unit_str)}"
        return stats


#=====================================================================================
#=====================================================================================
#  r e t i m e
#=====================================================================================
#=====================================================================================

def retime(time_sec, unitC):
    """
## retime (time_sec, unitC) - Convert time value in seconds to unitC resolution

`retime()` translates a value is resolution seconds into a new target resolution


### Args
`time_sec` (int or float)
- The time value in resolution seconds to be converted

`unitC` (str)
- Target time resolution: "s", "m", "h", "d", or "w" (case insensitive)


### Returns
- `time_sec` value scaled for the specified `unitC`, type float
- Raises ValueError if not given an int or float value for `time_sec`, or given an unsupported 
  unitC time unit suffix.
    """
    unitC = unitC.lower()
    if type(time_sec) in [int, float]:
        if unitC == "s":  return time_sec
        if unitC == "m":  return time_sec /60
        if unitC == "h":  return time_sec /60/60
        if unitC == "d":  return time_sec /60/60/24
        if unitC == "w":  return time_sec /60/60/24/7
        raise ValueError(f"Invalid unitC value <{unitC}> passed to retime()")
    else:
        raise ValueError(f"Invalid seconds value <{time_sec}> passed to retime().  Must be type int or float.")


#=====================================================================================
#=====================================================================================
#   g e t _ n e x t _ d t
#=====================================================================================
#=====================================================================================

str_to_daynum_map = {'monday':1, 'tuesday':2, 'wednesday':3, 'thursday':4, 'friday':5, 'saturday':6, 'sunday':7}

def get_next_dt(times, days=0, usec_resolution=False, test_dt=None):
    """
## get_next_dt (times, days=0, usec_resolution=False, test_dt=None) - Return next scheduled datetime occurrence

Given a list of days of the week and a list of clock times, return the next datetime after now.
The times arg may be a timevalue (eg '20m'), which allows user script code to accept/pass either a clock time or a 
time offset with the same syntax, simplifying the script code.


### Args
`times` (timevalue, or str or list of strs)
- If timevalue (eg, '20m'), then value is taken as a time offset from now, equivalent to `datetime.datetime.now() + timevalue`.
Note that using timevalue offsets may accumulate a time drift.
- If single time (eg, '17:20') or list of str times (eg ['12:13:14', '14:00']), then taken as specific times.
- times format is hour:minute with optional :second (default :00 seconds) using a 24-hour clock and local time zone.
- Note: A typo on a time replacing a ':' with '.' will be interpreted as a valid decimal number of seconds of a timevalue.

`days` (int/str or list of int/str, default 0)
- Single day or list of days.  Day numbers or string names accepted.
- Day numbers:  1 = Monday, 7 = Sunday (`isoweekday()` numbering)
- Day names accepted, case insensitive, eg 'MonDay'
- `days = 0` means every day, equiv [1,2,3,4,5,6,7]
- Alternately, if `days` is a str with a 'd' or 'w' suffix (case insensitive), the value is taken as a days or weeks 
offset from now at the earliest time in the `times` list.  The days/weeks offset value must >= 0 (today).  Examples:
  - Given `times = '08:00'` and `days = '3d'`, the returned datetime will be 3 days from today at 8AM
  - Given `times = ['12:34', '08:00']` and `days = '3w'`, the returned datetime will be 3 weeks from today at 8AM

`usec_resolution` (bool, default False)
- If True, timevalue offsets retain sub-second resolution
- If False, timevalue offsets are rounded to 1-second resolution
- When working with `times, days` values/lists the returned values are alway 1-second resolution

`test_dt` (datetime, default None)
- overrides datetime.datetime.now() for testing purposes


### Returns 
- datetime of next day/time match, or now+timevalue
- Raises ValueError with invalid input args
    """

    now_dt = datetime.datetime.now()  if test_dt is None  else test_dt


    # ----- timevalue case -----
    try:
        offset_sec = timevalue(times).seconds
        if not usec_resolution:
            return now_dt.replace(microsecond=0) + datetime.timedelta(seconds=offset_sec)
        else:
            return now_dt + datetime.timedelta(seconds=offset_sec)
    except:
        pass


    # ----- days/weeks offset case -----
    if isinstance(days, str)  and  (days.lower().endswith('d')  or  days.lower().endswith('w')):
        try:
            day_offset = int(days[:-1])
            if days.lower().endswith('w'):
                day_offset *= 7
        except Exception as e:
            raise ValueError (f"Invalid days arg <{days}>:  {e}")

        if day_offset < 0:
            raise ValueError (f"Invalid days arg <{days}>")

        _times_list = times
        if isinstance(_times_list, str):
            _times_list = [_times_list]

        _next_dt = now_dt + datetime.timedelta(day_offset)

        for tme in _times_list:
            try:
                _time = tme.split(':')
                if len(_time) in [2, 3]:
                    target_hour = int(_time[0])
                    target_minute = int(_time[1])
                    if len(_time) == 3:
                        target_second = int(_time[2])
                    else:
                        target_second = 0
                else:
                    raise ValueError (f"Invalid times arg <{tme}>")
            except:
                raise ValueError (f"Invalid times arg <{tme}>")

            with_time_dt = _next_dt.replace(hour=target_hour, minute=target_minute, second=target_second, microsecond=0)
            if with_time_dt < _next_dt:
                _next_dt = with_time_dt
        return _next_dt


    # ----- days/times lists case -----
    _days_list = days
    if isinstance(_days_list, int)  or  isinstance(_days_list, str):
        _days_list = [_days_list]
    elif not isinstance(_days_list, list):
        raise ValueError (f"Invalid days arg <{days}>")

    if len(_days_list) == 0:
        raise ValueError (f"Invalid days arg <{days}>")

    for n in range(len(_days_list)):                                    # Scrub the days list
        day_value = _days_list[n]
        if isinstance(day_value, int):
            if day_value < 0  or  day_value > 7:
                raise ValueError (f"Invalid day number <{day_value}>")
        elif isinstance(day_value, str):
            try:
                _days_list[n] = str_to_daynum_map[day_value.lower()]
            except:
                raise ValueError (f"Invalid day string <{day_value}>")
        else:
            raise ValueError (f"Invalid days arg <{days}>")

    today_daynum = now_dt.isoweekday()
    _next_dt = now_dt + datetime.timedelta(days=10)                     # start way out there and work back toward now_dt

    _times_list = times
    if isinstance(_times_list, str):
        _times_list = [_times_list]

    for daynum in _days_list:
        if daynum > 0:
            offset_num_days = daynum - today_daynum
            if offset_num_days < 0:
                offset_num_days += 7
        else:
            offset_num_days = 0
        plus_day_dt =  now_dt + datetime.timedelta(days=offset_num_days)

        for tme in _times_list:
            try:
                _time = tme.split(':')
                if len(_time) in [2, 3]:
                    target_hour = int(_time[0])
                    target_minute = int(_time[1])
                    if len(_time) == 3:
                        target_second = int(_time[2])
                    else:
                        target_second = 0
                else:
                    raise ValueError (f"Invalid times arg <{tme}>")
            except:
                raise ValueError (f"Invalid times arg <{tme}>")

            with_time_dt = plus_day_dt.replace(hour=target_hour, minute=target_minute, second=target_second, microsecond=0)
            if with_time_dt <= now_dt:
                if daynum == 0:
                    with_time_dt += datetime.timedelta(days=1)
                else:
                    with_time_dt += datetime.timedelta(days=7)
            if with_time_dt < _next_dt:
                _next_dt = with_time_dt
    return _next_dt
